#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.04 19:00:00                  #
# ================================================== #

import os
from typing import Optional, Dict, List

from pygpt_net.item.attachment import AttachmentItem
from pygpt_net.item.ctx import CtxItem


class Vision:
    def __init__(self, window=None):
        """
        Vision helpers for xAI (image inputs as data: URIs).

        :param window: Window instance
        """
        self.window = window
        self.attachments: Dict[str, str] = {}
        self.urls: List[str] = []
        self.input_tokens = 0
        self.allowed_mimes = {"image/jpeg", "image/png"}

    def build_images_for_chat(self, attachments: Optional[Dict[str, AttachmentItem]]) -> List[str]:
        """
        Build image inputs for xai_sdk.chat.image(...).
        Returns list of image sources (URLs or data: URIs).

        :param attachments: Attachments dict (id -> AttachmentItem)
        :return: List of image sources
        """
        import base64

        images: List[str] = []
        self.attachments = {}
        self.urls = []

        if not attachments:
            return images

        for id_, att in (attachments or {}).items():
            try:
                if att.path and self.window.core.api.xai.vision.is_image(att.path):
                    mime = self.window.core.api.xai.vision.guess_mime(att.path)
                    with open(att.path, "rb") as f:
                        b64 = base64.b64encode(f.read()).decode("utf-8")
                    images.append(f"data:{mime};base64,{b64}")
                    self.attachments[id_] = att.path
                    att.consumed = True
            except Exception:
                continue
        return images

    def is_image(self, path: str) -> bool:
        """
        Return True if path looks like an image file.

        :param path: File path
        """
        return path.lower().endswith(('.png', '.jpg', '.jpeg', '.tiff', '.bmp', '.gif', '.webp'))

    def guess_mime(self, path: str) -> str:
        """Guess MIME by extension.

        :param path: File path
        :return: MIME type string
        """
        ext = os.path.splitext(path)[1].lower().lstrip(".")
        if ext in ("jpg", "jpeg"):
            return "image/jpeg"
        if ext == "png":
            return "image/png"
        if ext == "gif":
            return "image/gif"
        if ext == "bmp":
            return "image/bmp"
        if ext == "webp":
            return "image/webp"
        if ext == "tiff":
            return "image/tiff"
        return "image/jpeg"

    def append_images(self, ctx: CtxItem):
        """
        Append sent images list to context for UI/history.

        :param ctx: CtxItem
        """
        images = self.get_attachments()
        if len(images) > 0:
            ctx.images = self.window.core.filesystem.make_local_list(list(images.values()))

    def get_attachments(self) -> Dict[str, str]:
        return self.attachments

    def get_urls(self) -> List[str]:
        return self.urls

    def reset_tokens(self):
        self.input_tokens = 0

    def get_used_tokens(self) -> int:
        return self.input_tokens

    def reset(self):
        self.attachments = {}
        self.urls = []
        self.input_tokens = 0


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.23 07:00:00                  #
# ================================================== #

import platform
import ctypes
from ctypes import wintypes
from typing import Dict, List, Optional, Tuple

# Guard for non-Windows platforms: provide stubs that fail at call-time
if platform.system() != "Windows":
    class WinAPI:
        def __init__(self): raise RuntimeError("Windows only")
    class InputSender:
        def __init__(self): raise RuntimeError("Windows only")
else:
    user32 = ctypes.windll.user32
    kernel32 = ctypes.windll.kernel32
    psapi = ctypes.windll.psapi
    gdi32 = ctypes.windll.gdi32

    # Constants
    GWL_EXSTYLE = -20
    WS_EX_LAYERED = 0x00080000
    LWA_ALPHA = 0x00000002

    SW_HIDE = 0
    SW_SHOWNORMAL = 1
    SW_SHOWMINIMIZED = 2
    SW_SHOWMAXIMIZED = 3
    SW_RESTORE = 9

    SWP_NOSIZE = 0x0001
    SWP_NOMOVE = 0x0002
    SWP_NOZORDER = 0x0004
    SWP_NOACTIVATE = 0x0010
    HWND_TOP = 0
    HWND_TOPMOST = -1
    HWND_NOTOPMOST = -2

    WM_CLOSE = 0x0010
    SMTO_BLOCK = 0x0001
    SMTO_ABORTIFHUNG = 0x0002
    SEND_TIMEOUT = 2000

    PROCESS_QUERY_INFORMATION = 0x0400
    PROCESS_VM_READ = 0x0010
    PROCESS_QUERY_LIMITED_INFORMATION = 0x1000

    # Input constants
    INPUT_MOUSE = 0
    INPUT_KEYBOARD = 1

    KEYEVENTF_KEYUP = 0x0002
    KEYEVENTF_UNICODE = 0x0004

    MOUSEEVENTF_MOVE = 0x0001
    MOUSEEVENTF_LEFTDOWN = 0x0002
    MOUSEEVENTF_LEFTUP = 0x0004
    MOUSEEVENTF_RIGHTDOWN = 0x0008
    MOUSEEVENTF_RIGHTUP = 0x0010
    MOUSEEVENTF_MIDDLEDOWN = 0x0020
    MOUSEEVENTF_MIDDLEUP = 0x0040

    # VK map
    VK = {
        "SHIFT": 0x10, "CTRL": 0x11, "ALT": 0x12, "WIN": 0x5B,
        "ENTER": 0x0D, "ESC": 0x1B, "TAB": 0x09, "SPACE": 0x20, "BACKSPACE": 0x08, "DELETE": 0x2E,
        "UP": 0x26, "DOWN": 0x28, "LEFT": 0x25, "RIGHT": 0x27,
        "HOME": 0x24, "END": 0x23, "PGUP": 0x21, "PGDN": 0x22,
        "F1": 0x70, "F2": 0x71, "F3": 0x72, "F4": 0x73, "F5": 0x74, "F6": 0x75,
        "F7": 0x76, "F8": 0x77, "F9": 0x78, "F10": 0x79, "F11": 0x7A, "F12": 0x7B,
        "F13": 0x7C, "F14": 0x7D, "F15": 0x7E, "F16": 0x7F, "F17": 0x80, "F18": 0x81,
        "F19": 0x82, "F20": 0x83, "F21": 0x84, "F22": 0x85, "F23": 0x86, "F24": 0x87,
    }

    if hasattr(wintypes, "ULONG_PTR"):
        ULONG_PTR = wintypes.ULONG_PTR
    else:
        ULONG_PTR = ctypes.c_ulonglong if ctypes.sizeof(ctypes.c_void_p) == 8 else ctypes.c_ulong

    if ctypes.sizeof(ctypes.c_void_p) == 8:
        LONG_PTR = ctypes.c_longlong
    else:
        LONG_PTR = ctypes.c_long

    # Structures
    class KEYBDINPUT(ctypes.Structure):
        _fields_ = [
            ("wVk", wintypes.WORD),
            ("wScan", wintypes.WORD),
            ("dwFlags", wintypes.DWORD),
            ("time", wintypes.DWORD),
            ("dwExtraInfo", ULONG_PTR),
        ]

    class MOUSEINPUT(ctypes.Structure):
        _fields_ = [
            ("dx", wintypes.LONG),
            ("dy", wintypes.LONG),
            ("mouseData", wintypes.DWORD),
            ("dwFlags", wintypes.DWORD),
            ("time", wintypes.DWORD),
            ("dwExtraInfo", ULONG_PTR),
        ]

    class HARDWAREINPUT(ctypes.Structure):
        _fields_ = [
            ("uMsg", wintypes.DWORD),
            ("wParamL", wintypes.WORD),
            ("wParamH", wintypes.WORD),
        ]

    class INPUT_union(ctypes.Union):
        _fields_ = [("ki", KEYBDINPUT), ("mi", MOUSEINPUT), ("hi", HARDWAREINPUT)]

    class INPUT(ctypes.Structure):
        _fields_ = [("type", wintypes.DWORD), ("union", INPUT_union)]

    # Function prototypes (partial)
    EnumWindowsProc = ctypes.WINFUNCTYPE(wintypes.BOOL, wintypes.HWND, wintypes.LPARAM)
    EnumChildProc = ctypes.WINFUNCTYPE(wintypes.BOOL, wintypes.HWND, wintypes.LPARAM)

    user32.EnumWindows.restype = wintypes.BOOL
    user32.EnumWindows.argtypes = (EnumWindowsProc, wintypes.LPARAM)
    user32.EnumChildWindows.restype = wintypes.BOOL
    user32.EnumChildWindows.argtypes = (wintypes.HWND, EnumChildProc, wintypes.LPARAM)

    user32.SendInput.restype = wintypes.UINT
    user32.SendInput.argtypes = (wintypes.UINT, ctypes.POINTER(INPUT), ctypes.c_int)

    # Helpers for Get/SetWindowLongPtr
    try:
        GetWindowLongPtrW = user32.GetWindowLongPtrW
        SetWindowLongPtrW = user32.SetWindowLongPtrW
        GetWindowLongPtrW.restype = LONG_PTR
        GetWindowLongPtrW.argtypes = (wintypes.HWND, ctypes.c_int)
        SetWindowLongPtrW.restype = LONG_PTR
        SetWindowLongPtrW.argtypes = (wintypes.HWND, ctypes.c_int, LONG_PTR)
        _use_long_ptr = True
    except AttributeError:
        GetWindowLongW = user32.GetWindowLongW
        SetWindowLongW = user32.SetWindowLongW
        GetWindowLongW.restype = ctypes.c_long
        GetWindowLongW.argtypes = (wintypes.HWND, ctypes.c_int)
        SetWindowLongW.restype = ctypes.c_long
        SetWindowLongW.argtypes = (wintypes.HWND, ctypes.c_int, ctypes.c_long)
        _use_long_ptr = False

    def _get_window_text(hwnd: int) -> str:
        buf = ctypes.create_unicode_buffer(512)
        user32.GetWindowTextW(wintypes.HWND(hwnd), buf, 512)
        return buf.value or ""

    def _get_class_name(hwnd: int) -> str:
        buf = ctypes.create_unicode_buffer(256)
        user32.GetClassNameW(wintypes.HWND(hwnd), buf, 256)
        return buf.value or ""

    def _get_pid_exe(hwnd: int) -> Tuple[int, Optional[str]]:
        pid = wintypes.DWORD()
        user32.GetWindowThreadProcessId(wintypes.HWND(hwnd), ctypes.byref(pid))
        exe_path = None
        access = PROCESS_QUERY_LIMITED_INFORMATION | PROCESS_VM_READ | PROCESS_QUERY_INFORMATION
        hProcess = kernel32.OpenProcess(access, False, pid.value)
        if hProcess:
            try:
                buf_len = wintypes.DWORD(4096)
                buf = ctypes.create_unicode_buffer(4096)
                QueryFullProcessImageNameW = getattr(kernel32, "QueryFullProcessImageNameW", None)
                if QueryFullProcessImageNameW:
                    if QueryFullProcessImageNameW(hProcess, 0, buf, ctypes.byref(buf_len)):
                        exe_path = buf.value
                if not exe_path:
                    GetModuleFileNameExW = getattr(psapi, "GetModuleFileNameExW", None)
                    if GetModuleFileNameExW:
                        if GetModuleFileNameExW(hProcess, 0, buf, 4096):
                            exe_path = buf.value
            finally:
                kernel32.CloseHandle(hProcess)
        return pid.value, exe_path

    def _get_rect(hwnd: int) -> Dict:
        rect = wintypes.RECT()
        user32.GetWindowRect(wintypes.HWND(hwnd), ctypes.byref(rect))
        width = rect.right - rect.left
        height = rect.bottom - rect.top
        return {"left": rect.left, "top": rect.top, "right": rect.right, "bottom": rect.bottom,
                "width": width, "height": height}

    def _is_visible(hwnd: int) -> bool:
        return bool(user32.IsWindowVisible(wintypes.HWND(hwnd)))

    def _is_minimized(hwnd: int) -> bool:
        return bool(user32.IsIconic(wintypes.HWND(hwnd)))

    class WinAPI:
        def enum_windows(self, visible_only: bool = True) -> List[Dict]:
            items: List[Dict] = []

            @EnumWindowsProc
            def callback(hWnd, lParam):
                try:
                    if visible_only and not _is_visible(hWnd):
                        return True
                    title = _get_window_text(hWnd)
                    cls = _get_class_name(hWnd)
                    pid, exe = _get_pid_exe(hWnd)
                    rect = _get_rect(hWnd)
                    minimized = _is_minimized(hWnd)
                    items.append({
                        "hwnd": int(hWnd),
                        "title": title,
                        "class_name": cls,
                        "pid": pid,
                        "exe": exe or "",
                        "rect": rect,
                        "is_visible": _is_visible(hWnd),
                        "is_minimized": minimized,
                    })
                except Exception:
                    return True
                return True

            user32.EnumWindows(callback, 0)
            return items

        def enum_child_windows(self, parent_hwnd: int) -> List[Dict]:
            items: List[Dict] = []

            @EnumChildProc
            def callback(hChild, lParam):
                try:
                    title = _get_window_text(hChild)
                    cls = _get_class_name(hChild)
                    pid, exe = _get_pid_exe(hChild)
                    rect = _get_rect(hChild)
                    items.append({
                        "hwnd": int(hChild),
                        "title": title,
                        "class_name": cls,
                        "pid": pid,
                        "exe": exe or "",
                        "rect": rect,
                        "is_visible": _is_visible(hChild),
                        "is_minimized": _is_minimized(hChild),
                    })
                except Exception:
                    return True
                return True

            user32.EnumChildWindows(wintypes.HWND(parent_hwnd), callback, 0)
            return items

        def is_window(self, hwnd: int) -> bool:
            return bool(user32.IsWindow(wintypes.HWND(hwnd)))

        def get_window_info(self, hwnd: int) -> Dict:
            return {
                "hwnd": int(hwnd),
                "title": _get_window_text(hwnd),
                "class_name": _get_class_name(hwnd),
                "pid": _get_pid_exe(hwnd)[0],
                "exe": _get_pid_exe(hwnd)[1] or "",
                "rect": _get_rect(hwnd),
                "is_visible": _is_visible(hwnd),
                "is_minimized": _is_minimized(hwnd),
            }

        def bring_to_foreground(self, hwnd: int) -> Tuple[bool, str]:
            if _is_minimized(hwnd):
                user32.ShowWindow(wintypes.HWND(hwnd), SW_RESTORE)
            if user32.SetForegroundWindow(wintypes.HWND(hwnd)):
                return True, "SetForegroundWindow succeeded."
            fg = user32.GetForegroundWindow()
            if fg:
                target_tid = user32.GetWindowThreadProcessId(wintypes.HWND(hwnd), None)
                fg_tid = user32.GetWindowThreadProcessId(wintypes.HWND(fg), None)
                cur_tid = kernel32.GetCurrentThreadId()
                user32.AttachThreadInput(fg_tid, cur_tid, True)
                user32.AttachThreadInput(target_tid, cur_tid, True)
                user32.BringWindowToTop(wintypes.HWND(hwnd))
                ok = user32.SetForegroundWindow(wintypes.HWND(hwnd))
                user32.AttachThreadInput(fg_tid, cur_tid, False)
                user32.AttachThreadInput(target_tid, cur_tid, False)
                if ok:
                    return True, "Foreground set via AttachThreadInput."
            return False, "Could not bring window to foreground (OS prevented focus change)."

        def move_resize(self, hwnd: int,
                        x: Optional[int], y: Optional[int],
                        width: Optional[int], height: Optional[int]) -> Tuple[bool, str]:
            rect = _get_rect(hwnd)
            new_x = rect["left"] if x is None else int(x)
            new_y = rect["top"] if y is None else int(y)
            new_w = rect["width"] if width is None else max(1, int(width))
            new_h = rect["height"] if height is None else max(1, int(height))
            ok = user32.SetWindowPos(wintypes.HWND(hwnd), HWND_TOP, new_x, new_y, new_w, new_h, 0)
            return (ok != 0), f"Moved to ({new_x},{new_y}) size ({new_w}x{new_h})."

        def show_window(self, hwnd: int, state: str) -> Tuple[bool, str]:
            m = {
                "minimize": SW_SHOWMINIMIZED,
                "maximize": SW_SHOWMAXIMIZED,
                "restore": SW_RESTORE,
                "show": SW_SHOWNORMAL,
                "hide": SW_HIDE,
            }
            if state not in m:
                return False, f"Invalid state: {state}"
            ok = user32.ShowWindow(wintypes.HWND(hwnd), m[state])
            return (ok != 0), f"ShowWindow {state}"

        def close_window(self, hwnd: int) -> Tuple[bool, str]:
            res = user32.SendMessageTimeoutW(
                wintypes.HWND(hwnd), WM_CLOSE, 0, 0,
                SMTO_BLOCK | SMTO_ABORTIFHUNG, SEND_TIMEOUT, None
            )
            return (res != 0), "WM_CLOSE sent."

        def set_topmost(self, hwnd: int, topmost: bool) -> Tuple[bool, str]:
            flag = HWND_TOPMOST if topmost else HWND_NOTOPMOST
            ok = user32.SetWindowPos(wintypes.HWND(hwnd), flag, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE)
            return (ok != 0), ("Topmost set" if topmost else "Topmost cleared")

        def _get_exstyle(self, hwnd: int) -> int:
            if '_use_long_ptr' in globals() and _use_long_ptr:
                return int(GetWindowLongPtrW(wintypes.HWND(hwnd), GWL_EXSTYLE))
            return int(GetWindowLongW(wintypes.HWND(hwnd), GWL_EXSTYLE))

        def _set_exstyle(self, hwnd: int, val: int) -> int:
            if '_use_long_ptr' in globals() and _use_long_ptr:
                return int(SetWindowLongPtrW(wintypes.HWND(hwnd), GWL_EXSTYLE, val))
            return int(SetWindowLongW(wintypes.HWND(hwnd), GWL_EXSTYLE, val))

        def set_opacity(self, hwnd: int, alpha: int) -> Tuple[bool, str]:
            ex = self._get_exstyle(hwnd)
            if (ex & WS_EX_LAYERED) == 0:
                self._set_exstyle(hwnd, ex | WS_EX_LAYERED)
            ok = user32.SetLayeredWindowAttributes(wintypes.HWND(hwnd), 0, wintypes.BYTE(alpha), LWA_ALPHA)
            return (ok != 0), f"Opacity set to alpha={alpha}"

        def get_foreground_window(self) -> Optional[int]:
            h = user32.GetForegroundWindow()
            return int(h) if h else None

        def get_window_rect(self, hwnd: int) -> Dict:
            return _get_rect(hwnd)

        def get_cursor_pos(self) -> Tuple[int, int]:
            pt = wintypes.POINT()
            user32.GetCursorPos(ctypes.byref(pt))
            return pt.x, pt.y

        def set_cursor_pos(self, x: int, y: int) -> bool:
            return bool(user32.SetCursorPos(int(x), int(y)))

    class InputSender:
        def __init__(self):
            pass

        def _send_input(self, inputs):
            n = len(inputs)
            arr = (INPUT * n)(*inputs)
            sent = user32.SendInput(n, arr, ctypes.sizeof(INPUT))
            return sent == n

        # ---- Keyboard ----
        def send_unicode_text(self, text: str, per_char_delay_ms: int = 2):
            for ch in text:
                code = ord(ch)
                down = INPUT(type=INPUT_KEYBOARD, union=INPUT_union(ki=KEYBDINPUT(wVk=0, wScan=code, dwFlags=KEYEVENTF_UNICODE, time=0, dwExtraInfo=0)))
                up = INPUT(type=INPUT_KEYBOARD, union=INPUT_union(ki=KEYBDINPUT(wVk=0, wScan=code, dwFlags=KEYEVENTF_UNICODE | KEYEVENTF_KEYUP, time=0, dwExtraInfo=0)))
                self._send_input([down, up])
                if per_char_delay_ms > 0:
                    kernel32.Sleep(per_char_delay_ms)

        def send_keys(self, keys: List[str], hold_ms: int = 50, gap_ms: int = 30) -> Tuple[bool, str]:
            tokens = [str(k).strip().upper() for k in keys if str(k).strip()]
            if not tokens:
                return False, "No key tokens."
            modifiers = [t for t in tokens if t in ("CTRL", "ALT", "SHIFT", "WIN")]
            normals = [t for t in tokens if t not in modifiers]

            def mk_down(vk):
                return INPUT(type=INPUT_KEYBOARD, union=INPUT_union(ki=KEYBDINPUT(wVk=vk, wScan=0, dwFlags=0, time=0, dwExtraInfo=0)))

            def mk_up(vk):
                return INPUT(type=INPUT_KEYBOARD, union=INPUT_union(ki=KEYBDINPUT(wVk=vk, wScan=0, dwFlags=KEYEVENTF_KEYUP, time=0, dwExtraInfo=0)))

            for mod in modifiers:
                vk = VK.get(mod)
                if vk is None: return False, f"Unsupported modifier: {mod}"
                if not self._send_input([mk_down(vk)]):
                    return False, f"Failed to press modifier: {mod}"

            for key in normals:
                vk = None
                if key in VK:
                    vk = VK[key]
                elif len(key) == 1:
                    c = key.upper()
                    if 'A' <= c <= 'Z' or '0' <= c <= '9':
                        vk = ord(c)
                if vk is None:
                    for mod in reversed(modifiers):
                        vkmod = VK.get(mod)
                        if vkmod is not None:
                            self._send_input([mk_up(vkmod)])
                    return False, f"Unsupported key token: {key}"

                if not self._send_input([mk_down(vk), mk_up(vk)]):
                    for mod in reversed(modifiers):
                        vkmod = VK.get(mod)
                        if vkmod is not None:
                            self._send_input([mk_up(vkmod)])
                    return False, f"Failed to tap: {key}"
                if gap_ms > 0:
                    kernel32.Sleep(gap_ms)

            for mod in reversed(modifiers):
                vk = VK.get(mod)
                if vk is not None:
                    self._send_input([mk_up(vk)])

            return True, "Keys sent."

        # ---- Mouse ----
        def _mouse_event(self, flags: int):
            mi = MOUSEINPUT(dx=0, dy=0, mouseData=0, dwFlags=flags, time=0, dwExtraInfo=0)
            inp = INPUT(type=INPUT_MOUSE, union=INPUT_union(mi=mi))
            return self._send_input([inp])

        def move_cursor(self, x: int, y: int) -> bool:
            return bool(user32.SetCursorPos(int(x), int(y)))

        def click_at(self, x: int, y: int, button: str = "left", double: bool = False) -> Tuple[bool, str]:
            self.move_cursor(x, y)
            btn = (button or "left").lower()
            if btn == "left":
                down, up = MOUSEEVENTF_LEFTDOWN, MOUSEEVENTF_LEFTUP
            elif btn == "right":
                down, up = MOUSEEVENTF_RIGHTDOWN, MOUSEEVENTF_RIGHTUP
            elif btn == "middle":
                down, up = MOUSEEVENTF_MIDDLEDOWN, MOUSEEVENTF_MIDDLEUP
            else:
                return False, f"Unsupported button: {button}"
            if not self._mouse_event(down) or not self._mouse_event(up):
                return False, "Mouse click failed."
            if double:
                kernel32.Sleep(80)
                if not self._mouse_event(down) or not self._mouse_event(up):
                    return False, "Mouse double-click failed."
            return True, "Click sent."

        def drag_and_drop(self, x1: int, y1: int, x2: int, y2: int, steps: int = 20, step_delay_ms: int = 10) -> Tuple[bool, str]:
            self.move_cursor(x1, y1)
            if not self._mouse_event(MOUSEEVENTF_LEFTDOWN):
                return False, "Mouse down failed."
            try:
                if steps < 1:
                    steps = 1
                dx = (x2 - x1) / float(steps)
                dy = (y2 - y1) / float(steps)
                cx, cy = float(x1), float(y1)
                for i in range(steps):
                    cx += dx
                    cy += dy
                    self.move_cursor(int(cx), int(cy))
                    if step_delay_ms > 0:
                        kernel32.Sleep(step_delay_ms)
            finally:
                self._mouse_event(MOUSEEVENTF_LEFTUP)
            return True, "Drag-and-drop completed."


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.24 23:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QWidget, QCheckBox

from pygpt_net.ui.widget.option.combo import OptionCombo
from pygpt_net.ui.widget.option.slider import OptionSlider
from pygpt_net.ui.widget.option.toggle_label import ToggleLabel
from pygpt_net.utils import trans


class AgentLlama:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: Window instance
        """
        self.window = window

    def setup(self) -> QWidget:
        """
        Setup agent llama options

        :return: QWidget
        """
        win = self.window
        ui = win.ui
        nodes = ui.nodes
        config_global = ui.config.setdefault('global', {})
        llama_opts = win.controller.agent.llama.options

        container = QWidget(win)

        nodes['agent.llama.loop.score'] = OptionSlider(
            win,
            'global',
            'agent.llama.loop.score',
            llama_opts["agent.llama.loop.score"],
        )
        nodes['agent.llama.loop.score'].setToolTip(trans("toolbox.agent.llama.loop.score.tooltip"))
        config_global['agent.llama.loop.score'] = nodes['agent.llama.loop.score']

        nodes['agent.llama.loop.mode.label'] = QLabel(trans("toolbox.agent.llama.loop.mode.label"), parent=win)
        nodes['agent.llama.loop.mode'] = OptionCombo(
            win,
            'global',
            'agent.llama.loop.mode',
            llama_opts["agent.llama.loop.mode"],
        )
        nodes['agent.llama.loop.mode'].setToolTip(trans("toolbox.agent.llama.loop.mode.tooltip"))
        config_global['agent.llama.loop.mode'] = nodes['agent.llama.loop.mode']

        nodes['agent.llama.loop.enabled'] = ToggleLabel(trans("toolbox.agent.llama.loop.enabled.label"), parent=win)
        nodes['agent.llama.loop.enabled'].box.toggled.connect(
            win.controller.agent.common.toggle_loop
        )
        config_global['agent.llama.loop.enabled'] = nodes['agent.llama.loop.enabled']

        nodes['agent.llama.loop.label'] = QLabel(trans("toolbox.agent.llama.loop.label"), parent=win)

        cols = QHBoxLayout()
        cols.addWidget(config_global['agent.llama.loop.enabled'])
        cols.addWidget(config_global['agent.llama.loop.score'])

        rows = QVBoxLayout()
        rows.addWidget(nodes['agent.llama.loop.label'])
        rows.addLayout(cols)
        rows.addWidget(nodes['agent.llama.loop.mode'])

        container.setLayout(rows)
        container.setContentsMargins(0, 0, 0, 0)

        nodes['agent_llama.options'] = container
        return nodes['agent_llama.options']


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.31 23:00:00                  #
# ================================================== #

from typing import Optional
from pygpt_net.item.ctx import CtxItem

def set_ctx_rt_handle(ctx: Optional[CtxItem], handle: Optional[str], window=None):
    """Persist server session handle into ctx.extra['rt_session_id'] (best effort)."""
    try:
        if not ctx:
            return
        if not isinstance(ctx.extra, dict):
            ctx.extra = {}
        val = (handle or "").strip()
        if val:
            ctx.extra["rt_session_id"] = val
            if window:
                try:
                    window.core.ctx.update_item(ctx)
                except Exception:
                    pass
    except Exception:
        pass

def set_rt_session_expires_at(ctx: Optional[CtxItem], epoch_seconds: Optional[int], window=None):
    """Persist optional session expiration timestamp into ctx.extra."""
    if not ctx or epoch_seconds is None:
        return
    try:
        if not isinstance(ctx.extra, dict):
            ctx.extra = {}
        ctx.extra["rt_session_expires_at"] = int(epoch_seconds)
        if window:
            try:
                window.core.ctx.update_item(ctx)
            except Exception:
                pass
    except Exception:
        pass

def extract_last_session_id(items: list[CtxItem]) -> Optional[str]:
    """Extract last known session ID from a list of CtxItems."""
    if not items:
        return None
    for item in reversed(items):
        if not item or not isinstance(item.extra, dict):
            continue
        val = item.extra.get("rt_session_id")
        if isinstance(val, str) and val.strip():
            return val.strip()
    return None


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.31 23:00:00                  #
# ================================================== #

import asyncio
import threading
from typing import Optional

class BackgroundLoop:
    """
    Dedicated background asyncio loop running in its own thread.
    Safe cross-thread scheduling and sync wrappers.
    """
    def __init__(self, name: str = "RT-Loop"):
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._thread: Optional[threading.Thread] = None
        self._name = name

    @property
    def loop(self) -> Optional[asyncio.AbstractEventLoop]:
        return self._loop

    def ensure(self):
        if self._loop and self._loop.is_running():
            return
        self._loop = asyncio.new_event_loop()

        def _runner(loop: asyncio.AbstractEventLoop):
            asyncio.set_event_loop(loop)
            loop.run_forever()

        self._thread = threading.Thread(target=_runner, args=(self._loop,), name=self._name, daemon=True)
        self._thread.start()

    async def run(self, coro):
        if not self._loop:
            raise RuntimeError("Owner loop is not running")
        cfut = asyncio.run_coroutine_threadsafe(coro, self._loop)
        return await asyncio.wrap_future(cfut)

    def run_sync(self, coro, timeout: float = 5.0):
        if not self._loop or not self._loop.is_running():
            return None
        fut = asyncio.run_coroutine_threadsafe(coro, self._loop)
        try:
            return fut.result(timeout=timeout)
        except Exception:
            return None

    def stop(self, timeout: float = 2.0):
        loop, thread = self._loop, self._thread
        if loop and loop.is_running():
            loop.call_soon_threadsafe(loop.stop)
        if thread and thread.is_alive():
            thread.join(timeout=timeout)
        self._loop = None
        self._thread = None


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.11 14:00:00                  #
# ================================================== #

from PySide6.QtCore import QRunnable, Slot, QObject, Signal


class Worker(QRunnable):
    def __init__(self, fn, *args, **kwargs):
        super().__init__()
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()

    @Slot()
    def run(self):
        try:
            self.fn(*self.args, **self.kwargs)
        except Exception as e:
            pass
        finally:
            self.cleanup()

    def cleanup(self):
        """Cleanup resources after worker execution."""
        sig = self.signals
        self.signals = None
        if sig is not None:
            try:
                sig.deleteLater()
            except RuntimeError:
                pass


class WorkerSignals(QObject):
    updated = Signal(object)
    finished = Signal(object)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.17 05:00:00                  #
# ================================================== #

import json

from pygpt_net.core.types import (
    MODE_COMPUTER,
    OPENAI_REMOTE_TOOL_DISABLE_CODE_INTERPRETER,
    OPENAI_REMOTE_TOOL_DISABLE_COMPUTER_USE,
    OPENAI_REMOTE_TOOL_DISABLE_IMAGE,
    OPENAI_REMOTE_TOOL_DISABLE_WEB_SEARCH,
    OPENAI_REMOTE_TOOL_DISABLE_FILE_SEARCH,
    OPENAI_REMOTE_TOOL_DISABLE_MCP,
)

from pygpt_net.item.model import ModelItem
from pygpt_net.item.preset import PresetItem
from pygpt_net.utils import trans


class RemoteTools:

    REMOTE_TOOLS = [
        "web_search",
        "image",
        "code_interpreter",
        "mcp",
        "file_search",
        "computer_use",
    ]

    def __init__(self, window=None):
        """
        Remote tools

        :param window: Window instance
        """
        self.window = window

    def get_choices(self):
        """
        Return remote tools choices

        :return: List of remote tools
        """
        choices = []
        for tool in self.REMOTE_TOOLS:
            choices.append({
                tool: trans(f"remote_tool.openai.{tool}")
            })
        return choices

    def append_to_tools(
            self,
            mode: str,
            model: ModelItem,
            stream: bool,
            is_expert_call: bool,
            tools: list,
            preset: PresetItem = None
    ) -> list:
        """
        Prepare remote tools for the model

        :param mode: Agent mode
        :param model: Model item
        :param stream: Streaming flag
        :param is_expert_call: Is expert call
        :param tools: List of current tools
        :param preset: Preset item (optional, used for expert calls)
        :return: List of tools
        """
        # disabled by default
        enabled = {
            "web_search": False,
            "image": False,
            "code_interpreter": False,
            "mcp": False,
            "file_search": False,
            "computer_use": False,
        }
        enabled_global = self.window.controller.chat.remote_tools.enabled  # get global config

        # from global config if not expert call
        if not is_expert_call:
            enabled["web_search"] = enabled_global(model, "web_search") # <-- from global config
            enabled["image"] = self.window.core.config.get("remote_tools.image", False)
            enabled["code_interpreter"] = self.window.core.config.get("remote_tools.code_interpreter", False)
            enabled["mcp"] = self.window.core.config.get("remote_tools.mcp", False)
            enabled["file_search"] = self.window.core.config.get("remote_tools.file_search", False)
            enabled["computer_use"] = (mode == MODE_COMPUTER or model.id.startswith("computer-use"))
        else:
            # for expert call, get from preset config
            if preset:
                if preset.remote_tools:
                    tools_list = [preset_remote_tool.strip() for preset_remote_tool in preset.remote_tools.split(",") if preset_remote_tool.strip()]
                    if "web_search" not in tools_list:
                        enabled["web_search"] = enabled_global(model, "web_search") # <-- from global config
                    for item in tools_list:
                        if item in enabled:
                            enabled[item] = True

        # o1, o3, models do not support remote tools
        # TODO: check if really not supported
        if model.id.startswith("o1") or model.id.startswith("o3"):
            return tools

        # extend local tools with remote tools
        if enabled["computer_use"]:
            if not model.id in OPENAI_REMOTE_TOOL_DISABLE_COMPUTER_USE:
                tools.append(self.window.core.api.openai.computer.get_tool())
        else:
            if not model.id in OPENAI_REMOTE_TOOL_DISABLE_WEB_SEARCH:
                if enabled["web_search"]:
                    tools.append({"type": "web_search_preview"})

            if not model.id in OPENAI_REMOTE_TOOL_DISABLE_CODE_INTERPRETER:
                if enabled["code_interpreter"]:
                    tools.append({
                        "type": "code_interpreter",
                        "container": {
                            "type": "auto"
                        }
                    })

            if not model.id in OPENAI_REMOTE_TOOL_DISABLE_IMAGE:
                if enabled["image"]:
                    tool = {"type": "image_generation"}
                    if stream:
                        tool["partial_images"] = 1  # required for streaming
                    tools.append(tool)

            if not model.id in OPENAI_REMOTE_TOOL_DISABLE_FILE_SEARCH:
                if enabled["file_search"]:
                    vector_store_ids = self.window.core.config.get("remote_tools.file_search.args", "")
                    if vector_store_ids:
                        vector_store_ids = [store.strip() for store in vector_store_ids.split(",") if store.strip()]
                        tools.append({
                            "type": "file_search",
                            "vector_store_ids": vector_store_ids,
                        })

            if not model.id in OPENAI_REMOTE_TOOL_DISABLE_MCP:
                if enabled["mcp"]:
                    mcp_tool = self.window.core.config.get("remote_tools.mcp.args", "")
                    if mcp_tool:
                        mcp_tool = json.loads(mcp_tool)
                        tools.append(mcp_tool)

        return tools


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.02 20:00:00                  #
# ================================================== #

from .openai import Store as OpenAIStore
from .google import Store as GoogleStore
from .anthropic import Store as AnthropicStore
from .xai import Store as XAIStore

class RemoteStore:
    def __init__(self, window=None):
        """
        Remote vector stores core

        :param window: Window instance
        """
        self.window = window
        self.openai = OpenAIStore(self.window)
        self.google = GoogleStore(self.window)
        self.anthropic = AnthropicStore(self.window)
        self.xai = XAIStore(self.window)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.28 09:00:00                  #
# ================================================== #

import json
from typing import Dict, List, Optional

from PySide6.QtCore import Slot

from pygpt_net.core.experts.worker import ExpertWorker
from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_EXPERT,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_AUDIO,
    MODE_RESEARCH,
    TOOL_EXPERT_CALL_NAME,
    TOOL_EXPERT_CALL_DESCRIPTION,
    TOOL_EXPERT_CALL_PARAM_ID_DESCRIPTION,
    TOOL_EXPERT_CALL_PARAM_QUERY_DESCRIPTION,
    TOOL_QUERY_ENGINE_NAME,
    TOOL_QUERY_ENGINE_DESCRIPTION,
    TOOL_QUERY_ENGINE_PARAM_QUERY_DESCRIPTION,
)
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.events import Event, KernelEvent
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.preset import PresetItem
from pygpt_net.utils import trans

class Experts:
    def __init__(self, window=None):
        """
        Experts core

        :param window: Window instance
        """
        self.window = window
        self.allowed_modes = [
            MODE_CHAT,
            MODE_COMPLETION,
            # MODE_VISION,
            # MODE_LANGCHAIN,
            MODE_LLAMA_INDEX,
            MODE_AUDIO,
            MODE_RESEARCH,
        ]
        self.allowed_cmds = [TOOL_EXPERT_CALL_NAME]
        self.worker = None
        self.last_expert_id = None  # last expert id used in call
        self.last_idx = None  # last index used in call
        self.master_ctx = None  # master meta for expert calls

    def get_mode(self) -> str:
        """
        Get sub-mode to use internally

        :return: sub-mode
        """
        mode = MODE_CHAT
        current = self.window.core.config.get("experts.mode")
        if current is not None and current in self.allowed_modes:
            mode = current
        return mode

    def stopped(self) -> bool:
        """
        Check if experts are stopped

        :return: True if stopped
        """
        return self.window.controller.kernel.stopped()

    def agent_enabled(self) -> bool:
        """
        Check if agent is enabled

        :return: True if enabled
        """
        return self.window.controller.agent.legacy.enabled()

    def exists(self, id: str) -> bool:
        """
        Check if expert exists

        :param id: expert id
        :return: True if exists
        """
        return self.window.core.presets.has(MODE_EXPERT, id)

    def get_expert(self, id: str) -> PresetItem:
        """
        Get expert by id

        :param id: expert id
        :return: expert item (preset)
        """
        return self.window.core.presets.get_by_id(MODE_EXPERT, id)

    def get_experts(self) -> Dict[str, PresetItem]:
        """
        Get experts names with keys

        :return: experts dict
        """
        experts = {}
        core = self.window.core
        presets = core.presets.get_by_mode(MODE_EXPERT)

        # mode: agent
        if self.agent_enabled():
            agents = core.presets.get_by_mode(MODE_AGENT)
            agent = core.config.get('preset')
            if agent is not None:
                if agent in agents:
                    for uuid in agents[agent].experts:
                        expert = core.presets.get_by_uuid(uuid)
                        if expert is not None:
                            id = expert.filename
                            experts[id] = expert
        # mode: expert
        else:
            for k in presets:
                if k.startswith("current."):  # skip current
                    continue
                if not presets[k].enabled:  # skip disabled experts
                    continue
                experts[k] = presets[k]
        return experts

    def get_expert_name_by_id(self, id: str) -> Optional[str]:
        """
        Get expert name by id

        :param id: expert id
        :return: expert name or None if not found
        """
        experts = self.get_experts()
        if id in experts:
            return experts[id].name

    def count_experts(self, uuid: str) -> int:
        """
        Count experts in agent

        :param uuid: agent uuid
        :return: number of experts
        """
        i = 0
        core = self.window.core
        agents = core.presets.get_by_mode(MODE_AGENT)
        if uuid in agents:
            for expert_uuid in agents[uuid].experts:
                expert = core.presets.get_by_uuid(expert_uuid)
                if expert is not None:
                    i += 1
        return i

    def get_prompt(self) -> str:
        """
        Get prompt for handle experts

        :return: prompt with list of experts appended
        """
        prompt = self.window.core.config.get('prompt.expert')
        experts = self.get_experts()
        experts_list = []
        for k in experts:
            if k.startswith("current."):  # skip current presets
                continue
            if experts[k].description.strip() == "":
                experts_list.append(f" - {k}: {experts[k].name}")
            else:
                experts_list.append(f" - {k}: {experts[k].name} ({experts[k].description})")
        return prompt.replace("{presets}", "\n".join(experts_list))

    def extract_calls(self, ctx: CtxItem) -> Dict[str, str]:
        """
        Extract expert calls from context output

        :param ctx: context item
        :return: dict with calls
        """
        core = self.window.core
        ids = self.get_experts().keys()
        if not ids:  # abort if no experts
            return {}
        cmds = core.command.extract_cmds(ctx.output)
        if len(cmds) > 0:
            ctx.cmds = cmds  # append commands to ctx
        else:  # abort if no cmds
            return {}
        commands = core.command.from_commands(cmds)  # pack to execution list
        is_cmd = False
        my_commands = []
        calls = {}
        for item in commands:
            if item["cmd"] in self.allowed_cmds:
                my_commands.append(item)
                is_cmd = True
        if not is_cmd:  # abort if no expert calls
            return {}
        for item in my_commands:
            try:
                if item["cmd"] == TOOL_EXPERT_CALL_NAME:
                    if "params" not in item:
                        continue
                    if "id" not in item["params"] or "query" not in item["params"]:
                        continue
                    id = item["params"]["id"]
                    if id not in ids:
                        continue
                    query = item["params"]["query"]
                    calls[id] = query
            except Exception as e:
                core.debug.log(e)
                return {}
        return calls

    def extract_tool_calls(self, ctx: CtxItem):
        """
        Extract tool calls from expert

        :param ctx: context item
        """
        for call in ctx.tool_calls:
            if (call["type"] == "function"
                    and "function" in call
                    and call["function"]["name"] == TOOL_QUERY_ENGINE_NAME):
                ctx.force_call = True  # force call if query engine tool is used
                ctx.cmds_before = [
                    {
                        "cmd": TOOL_QUERY_ENGINE_NAME,
                        "params": {
                            "query": call["function"]["arguments"]["query"],
                            "idx": self.last_idx,
                        },
                    }
                ]
                break

    def reply(self, ctx: CtxItem):
        """
        Re-send response from commands to master expert

        If command has been called by expert then response for command is send here to parent

        :param ctx: context item
        """
        if self.stopped():
            return

        # master meta is here, ctx.meta = MASTER META ID here!

        # make copy of ctx for reply, and change input name to expert name
        reply_ctx = CtxItem()
        reply_ctx.from_dict(ctx.to_dict())
        expert_id = self.last_expert_id
        expert_name = ""
        if expert_id:
            expert_name = self.get_expert_name_by_id(expert_id)

        # reply ctx has no meta here!!!!
        reply_ctx.input_name = expert_name
        reply_ctx.output_name = ""
        reply_ctx.sub_call = True  # this flag is not copied in to_dict

        if ctx.output.strip() != "":
            response = reply_ctx.output
        else:
            response = reply_ctx.input

        context = BridgeContext()
        context.ctx = reply_ctx
        result = {
            "expert_id": expert_id,
            "result": str(response),
        }
        context.prompt = str(response)
        if ctx.sub_reply:
            reply_ctx.extra["sub_reply"] = True  # mark as sub-reply
            context.prompt = json.dumps(result, ensure_ascii=False, indent=2)  # to master

        self.window.dispatch(KernelEvent(KernelEvent.INPUT_SYSTEM, {
            'context': context,
            'extra': {
                "force": True,
                "reply": True,
                "internal": self.agent_enabled(),
            },
        }))

    def call(
            self,
            master_ctx: CtxItem,
            expert_id: str,
            query: str
    ):
        """
        Call the expert

        :param master_ctx: master context
        :param expert_id: expert id (preset ID)
        :param query: input text (master prompt)
        """
        if self.stopped():
            return

        worker = ExpertWorker(
            window=self.window,
            master_ctx=master_ctx,
            expert_id=expert_id,
            query=query,
        )
        worker.signals.response.connect(self.handle_response)  # connect to finished signal
        worker.signals.finished.connect(self.handle_finished)  # connect to finished signal
        worker.signals.error.connect(self.handle_error)  # connect to error signal
        worker.signals.event.connect(self.handle_event)  # connect to event signal
        worker.signals.output.connect(self.handle_output)  # connect to output signal
        worker.signals.lock_input.connect(self.handle_input_locked)  # connect to lock input signal
        worker.signals.cmd.connect(self.handle_cmd)  # connect to cmd signal

        # start worker in threadpool
        self.worker = worker
        self.last_expert_id = expert_id  # store last expert id
        self.master_ctx = master_ctx
        name = self.get_expert_name_by_id(expert_id)
        event = KernelEvent(KernelEvent.STATE_BUSY, {
            "msg": f"{trans('expert.wait.status')} ({name})",
        })
        self.window.dispatch(event)  # dispatch busy event
        self.window.threadpool.start(self.worker)

    @Slot(CtxItem, str)
    def handle_output(self, ctx: CtxItem, mode: str):
        """
        Handle output from worker

        :param ctx: CtxItem
        :param mode: mode of operation
        """
        if self.stopped():
            return

        self.window.controller.chat.output.handle(
            ctx=ctx,
            mode=mode,
            stream=False,
        )

    @Slot(CtxItem, CtxItem, str, str, str)
    def handle_cmd(
            self,
            ctx: CtxItem,
            master_ctx: CtxItem,
            expert_id: str,
            expert_name: str,
            result: str
    ):
        """
        Handle command from worker

        :param ctx: CtxItem
        :param master_ctx: master context item
        :param expert_id: expert id
        :param expert_name: expert name
        :param result: result of command execution
        """
        if self.stopped():
            return

        core = self.window.core
        update_status = self.window.update_status

        # extract native tool calls if provided
        if ctx.tool_calls:
            # if not internal commands in a text body then append tool calls as commands (prevent double commands)
            if not core.command.has_cmds(ctx.output):
                core.command.append_tool_calls(ctx)  # append tool calls as commands
                if not isinstance(ctx.extra, dict):
                    ctx.extra = {}
                ctx.extra["tool_calls"] = ctx.tool_calls

        # if 'get_context' tool is used then force call, and append idx
        self.extract_tool_calls(ctx)  # extract tool calls from ctx
        self.window.controller.chat.command.handle(ctx, internal=True)  # handle cmds sync

        if ctx.reply:
            update_status("")  # clear status
            # prepare data to send as reply
            tool_data = json.dumps(ctx.results)
            # if "tool_output" in ctx.extra and ctx.extra["tool_output"]:
               # tool_data = str(ctx.extra["tool_output"])

            core.ctx.update_item(ctx)  # update context in db
            update_status('...')
            ctx.output = f"<tool>{ctx.cmds}</tool>"
            core.ctx.update_item(ctx)  # update ctx in DB
            self.handle_finished()
            self.call(
                master_ctx=self.master_ctx,
                expert_id=self.last_expert_id,
                query=tool_data,
            )
            return

        # ----- no cmd reply ------ #

        # make copy of ctx for reply, and change input name to expert name
        reply_ctx = CtxItem()
        reply_ctx.from_dict(ctx.to_dict())
        reply_ctx.meta = master_ctx.meta

        # assign expert output
        reply_ctx.output = result
        reply_ctx.input_name = expert_name
        reply_ctx.output_name = ""
        reply_ctx.cmds = []  # clear cmds
        reply_ctx.sub_call = True  # this flag is not copied in to_dict

        # reply to main thread
        # send to reply()
        # input: something (no tool results here)
        # output: ... (call the master)
        self.handle_response(reply_ctx, str(expert_id))

    @Slot()
    def handle_input_locked(self):
        """
        Handle input locked signal from worker

        This is used to lock input in UI while worker is processing.
        """
        if self.stopped():
            return
        self.window.controller.chat.common.lock_input()

    @Slot(Event)
    def handle_event(self, event: Event):
        """
        Handle event from worker

        :param event: Event instance
        """
        if self.stopped():
            return
        self.window.dispatch(event)

    @Slot(str)
    def handle_error(self, error: str):
        """
        Handle worker error

        :param error: error message
        """
        dispatch = self.window.dispatch

        if self.stopped():
            dispatch(KernelEvent(KernelEvent.STATE_IDLE, {}))  # dispatch idle event
            return

        # handle error from worker
        context = BridgeContext()
        context.prompt = f"{trans('expert.wait.failed')}: {error}"
        dispatch(KernelEvent(KernelEvent.INPUT_SYSTEM, {
            'context': context,
            'extra': {
                "force": True,
                "reply": False,
                "internal": False,
            },
        }))  # reply to master
        dispatch(KernelEvent(KernelEvent.STATE_IDLE, {}))  # dispatch idle event

    @Slot()
    def handle_finished(self):
        """Handle worker finished signal"""
        self.window.dispatch(KernelEvent(KernelEvent.STATE_IDLE, {}))  # dispatch idle event

    @Slot(CtxItem, str)
    def handle_response(self, ctx: CtxItem, expert_id: str):
        """
        Handle response expert call

        :param ctx: CtxItem
        :param expert_id: expert id
        """
        dispatch = self.window.dispatch

        if self.stopped():
            dispatch(KernelEvent(KernelEvent.STATE_IDLE, {}))  # dispatch idle event
            return

        # handle reply from worker
        context = BridgeContext()
        context.ctx = ctx
        result = {
            "expert_id": expert_id,
            "result": str(ctx.output),
        }
        # TODO: clear ctx.output here?
        context.prompt = json.dumps(result, ensure_ascii=False, indent=2)  # prepare prompt for reply
        dispatch(KernelEvent(KernelEvent.INPUT_SYSTEM, {
            'context': context,
            'extra': {
                "force": True,
                "reply": True,
                "internal": False,
            },
        }))  # reply to master
        dispatch(KernelEvent(KernelEvent.STATE_IDLE, {}))  # dispatch idle event

    def get_functions(self) -> List[Dict[str, str]]:
        """
        Append call the expert commands

        :return: call the expert commands
        """
        return [
            {
                "cmd": TOOL_EXPERT_CALL_NAME,
                "instruction": TOOL_EXPERT_CALL_DESCRIPTION,
                "params": [
                    {
                        "name": "id",
                        "description": TOOL_EXPERT_CALL_PARAM_ID_DESCRIPTION,
                        "required": True,
                        "type": "str",
                    },
                    {
                        "name": "query",
                        "description": TOOL_EXPERT_CALL_PARAM_QUERY_DESCRIPTION,
                        "required": True,
                        "type": "str",
                    }
                ]
            }
        ]

    def get_retriever_tool(self) -> Dict[str, str]:
        """
        Get retriever tool for additional context retrieval

        :return: retriever tool definition
        """
        return {
            "cmd": TOOL_QUERY_ENGINE_NAME,
            "instruction": TOOL_QUERY_ENGINE_DESCRIPTION,
            "params": [
                {
                    "name": "query",
                    "description": TOOL_QUERY_ENGINE_PARAM_QUERY_DESCRIPTION,
                    "required": True,
                    "type": "str",
                }
            ]
        }

    def has_calls(self, ctx: CtxItem) -> bool:
        """
        Check if context has expert calls

        :param ctx: CtxItem
        :return: True if expert calls found
        """
        if not ctx.sub_reply and not ctx.reply:
            mentions = self.extract_calls(ctx)
            if mentions:
                for expert_id in mentions:
                    if not self.exists(expert_id):
                        continue
                    return True
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.05 20:00:00                  #
# ================================================== #

import json
from typing import List, Any, Dict, Optional

from pygpt_net.item.model import ModelItem


class Tools:
    def __init__(self, window=None):
        """
        Tools mapper for Anthropic Messages API.

        :param window: Window instance
        """
        self.window = window

    def _sanitize_schema(self, schema: Any) -> Any:
        """
        Sanitize JSON Schema dict for Anthropic input_schema.

        - Remove unsupported or risky keywords.
        - Normalize 'type'.
        - Ensure properties/items recursively valid.

        :param schema: JSON Schema (dict or list)
        :return: Sanitized JSON Schema (dict)
        """
        # 1) entry point: if list, take the first element
        if isinstance(schema, list):
            # If it's a list of schemas/types, take the first one (after filtering out empty ones)
            return self._sanitize_schema(schema[0]) if schema else {}

        if not isinstance(schema, dict):
            return schema

        # 2) remove unsupported keys
        banned = {
            "unevaluatedProperties",
            "$defs", "$ref", "$schema", "$id",
            "examples", "readOnly", "writeOnly", "nullable",
            "dependentSchemas", "dependentRequired",
            "oneOf", "anyOf", "allOf", "patternProperties", "dependencies",
            "additional_properties",  # underscore
            "additionalProperties",  # camelCase
        }
        for k in list(schema.keys()):
            if k in banned:
                schema.pop(k, None)

        # 3) normalize 'type'
        t = schema.get("type")

        # a) list of types -> take the first non-null
        if isinstance(t, list):
            t_no_null = [x for x in t if isinstance(x, str) and x.lower() != "null"]
            schema["type"] = t_no_null[0] if t_no_null else "object"
            t = schema["type"]

        # b) if 'type' is not a string (e.g., dict), try to infer or remove it
        if not isinstance(t, str):
            if "properties" in schema:
                schema["type"] = "object"
            elif "items" in schema:
                schema["type"] = "array"
            elif "enum" in schema and isinstance(schema["enum"], list) and all(
                    isinstance(x, str) for x in schema["enum"]):
                schema["type"] = "string"
            else:
                # no reasonable type — leave without 'type' and continue
                schema.pop("type", None)
        else:
            schema["type"] = t.lower()

        # Safe form of type for further comparisons
        t_val = schema.get("type")
        type_l = t_val.lower() if isinstance(t_val, str) else ""

        # 4) enum only for string
        if "enum" in schema and type_l != "string":
            schema.pop("enum", None)

        # 5) Object
        if type_l == "object":
            props = schema.get("properties")
            if not isinstance(props, dict):
                props = {}
            clean_props: Dict[str, Any] = {}
            for pname, pval in props.items():
                clean_props[pname] = self._sanitize_schema(pval)
            schema["properties"] = clean_props

            req = schema.get("required")
            if not (isinstance(req, list) and all(isinstance(x, str) for x in req) and len(req) > 0):
                schema.pop("required", None)

        # 6) Array
        elif type_l == "array":
            items = schema.get("items")
            if isinstance(items, list):
                items = items[0] if items else {"type": "string"}
            if not isinstance(items, dict):
                items = {"type": "string"}
            schema["items"] = self._sanitize_schema(items)

        # 7) Recursion over remaining nestings,
        #    but skip 'properties' and 'items' — we've already sanitized them
        for k, v in list(schema.items()):
            if k in ("properties", "items"):
                continue
            if isinstance(v, dict):
                schema[k] = self._sanitize_schema(v)
            elif isinstance(v, list):
                schema[k] = [self._sanitize_schema(x) for x in v]

        return schema

    def prepare(self, model: ModelItem, functions: list) -> List[dict]:
        """
        Prepare Anthropic tool definitions: [{"name","description","input_schema"}].

        :param model: ModelItem
        :param functions: List of app function dicts
        :return: List of tool dicts for Anthropic
        """
        if not functions or not isinstance(functions, list):
            return []

        tools: List[dict] = []
        for fn in functions:
            name = str(fn.get("name") or "").strip()
            if not name:
                continue
            desc = fn.get("desc") or ""

            params: Optional[dict] = {}
            if fn.get("params"):
                try:
                    params = json.loads(fn["params"])
                except Exception:
                    params = {}
                params = self._sanitize_schema(params or {})
                if not params.get("type"):
                    params["type"] = "object"

            # pass through tool as client tool
            tool_def = {
                "name": name,
                "description": desc,
                "input_schema": params or {"type": "object"},
            }

            # optional: allow defer_loading for tool search when configured per-tool (kept compatible)
            if isinstance(fn, dict) and fn.get("defer_loading") is True:
                tool_def["defer_loading"] = True

            tools.append(tool_def)

        return tools

    def merge_tools_dedup(self, primary: List[dict], secondary: List[dict]) -> List[dict]:
        """
        Remove duplicate tools, preserving order:

        - First from primary list
        - Then from secondary list if not already present

        Dedup rules:
          * Tools with a 'name' are deduped by name.
          * MCP toolsets (type == 'mcp_toolset') are deduped by (type, mcp_server_name).
          * Tools without a 'name' use (type) as a fallback key.

        :param primary: Primary list of tool dicts
        :param secondary: Secondary list of tool dicts
        :return: Merged list of tool dicts without duplicates
        """
        def key_for(t: dict) -> str:
            name = t.get("name")
            if name:
                return f"name::{name}"
            ttype = t.get("type")
            if ttype == "mcp_toolset":
                return f"mcp::{t.get('mcp_server_name', '')}"
            return f"type::{ttype}"

        result: List[dict] = []
        seen = set()

        for t in primary or []:
            k = key_for(t)
            if k not in seen:
                seen.add(k)
                result.append(t)

        for t in secondary or []:
            k = key_for(t)
            if k in seen:
                continue
            seen.add(k)
            result.append(t)

        return result

    def get_all_tools(self, model: ModelItem, functions: list) -> List[dict]:
        """
        Get combined list of all tools (app functions + remote tools) for Anthropic.

        :param model: ModelItem
        :param functions: List of app function dicts
        :return: Combined list of tool dicts
        """
        base_tools = self.prepare(model, functions)
        remote_tools = self.window.core.api.anthropic.remote_tools.build_remote_tools(model)
        return self.merge_tools_dedup(base_tools, remote_tools)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.02 20:00:00                  #
# ================================================== #

from typing import Optional, List, Dict, Union

from packaging.version import Version

from pygpt_net.item.assistant import AssistantItem
from pygpt_net.item.store import RemoteFileItem
from pygpt_net.provider.core.remote_file.db_sqlite import DbSqliteProvider


class Files:

    PROVIDER_NAME = "google"

    def __init__(self, window=None):
        """
        Google remote files core (File Search documents)

        :param window: Window instance
        """
        self.window = window
        self.provider = DbSqliteProvider(window)
        self.items = {}

    def install(self):
        self.provider.install()

    def patch(self, app_version: Version) -> bool:
        return self.provider.patch(app_version)

    def get(self, id: str) -> Optional[RemoteFileItem]:
        if id in self.items:
            return self.items[id]
        return None

    def get_ids(self) -> List[str]:
        return list(self.items.keys())

    def get_all(self) -> Dict[str, RemoteFileItem]:
        return self.items

    def get_id_by_idx_all(self, idx: int) -> str:
        return list(self.items.keys())[idx]

    def get_by_idx(self, idx: int) -> str:
        items = self.items
        return list(items.keys())[idx]

    def has(self, id: str) -> bool:
        return id in self.items

    def create(
            self,
            assistant: AssistantItem,
            thread_id: str,
            file_id: str,
            name: str,
            path: str,
            size: int) -> Optional[RemoteFileItem]:
        """
        Not used in Google path (kept for parity). Use insert() instead.
        """
        file = RemoteFileItem()
        file.id = file_id
        file.file_id = file_id
        file.thread_id = thread_id
        file.provider = self.PROVIDER_NAME
        file.name = name
        file.path = path
        file.size = size
        if assistant.vector_store is not None and assistant.vector_store != "":
            file.store_id = assistant.vector_store
        file.record_id = self.provider.create(file)
        self.items[file.id] = file
        return file

    def update(self, file: RemoteFileItem) -> Optional[RemoteFileItem]:
        self.items[file.id] = file
        self.provider.save(file)
        return file

    def get_names(self) -> Dict[str, str]:
        names = {}
        for id in self.items:
            file = self.items[id]
            names[id] = file.name
        return names

    def get_by_store_or_thread(self, store_id: str, thread_id: str) -> Dict[str, RemoteFileItem]:
        return self.provider.get_by_store_or_thread(store_id, thread_id)

    def count_by_store_or_thread(self, store_id: str, thread_id: str) -> int:
        return self.provider.count_by_store_or_thread(store_id, thread_id)

    def get_file_by_idx(self, idx: int, store_id: str, thread_id: str) -> Optional[RemoteFileItem]:
        files = self.get_by_store_or_thread(store_id, thread_id)
        if idx >= len(files):
            return None
        return list(files.values())[idx]

    def get_file_id_by_idx(self, idx: int, store_id: str, thread_id: str) -> Optional[str]:
        file = self.get_file_by_idx(idx, store_id, thread_id)
        if file is None:
            return None
        return file.file_id

    def get_all_by_file_id(self, file_id: str) -> dict:
        return self.provider.get_all_by_file_id(file_id)

    def delete(self, file: Union[RemoteFileItem, list]) -> bool:
        files = file if isinstance(file, list) else [file]
        for f in files:
            file_id = f.file_id  # document name (fileSearchStores/.../documents/..)
            items = self.get_all_by_file_id(file_id)
            for id in items:
                store_id = items[id].store_id
                if store_id is None or store_id == "":
                    continue
                try:
                    self.window.core.api.google.store.delete_store_file(store_id, file_id)
                except Exception as e:
                    self.window.core.debug.log("Failed to delete document from store: " + str(e))
            self.provider.delete_by_id(f.record_id)
            try:
                # There is no Files API deletion here; this is store document
                pass
            except Exception as e:
                self.window.core.debug.log("Failed to delete remote document: " + str(e))
            if f.record_id in self.items:
                del self.items[f.record_id]
        return True

    def delete_by_file_id(self, file_id: str) -> bool:
        res = self.provider.delete_by_file_id(file_id)
        if res:
            to_delete = []
            for id in self.items:
                if self.items[id].file_id == file_id:
                    to_delete.append(id)
            for id in to_delete:
                del self.items[id]
        return res

    def on_store_deleted(self, store_id: str):
        self.provider.clear_store_from_files(store_id)

    def on_all_stores_deleted(self):
        self.provider.clear_all_stores_from_files(self.PROVIDER_NAME)

    def rename(self, record_id: int, name: str) -> bool:
        self.provider.rename_file(record_id, name)
        if record_id in self.items:
            self.items[record_id].name = name
        return True

    def truncate(self, store_id: Optional[str] = None) -> bool:
        if store_id is not None:
            self.window.core.api.google.store.remove_from_store(store_id)
        else:
            self.window.core.api.google.store.remove_from_stores()
        return self.truncate_local(store_id)

    def truncate_local(self, store_id: Optional[str] = None) -> bool:
        if store_id is not None:
            self.provider.truncate_by_store(store_id)
        else:
            self.provider.truncate_all(self.PROVIDER_NAME)
            self.items = {}
        return True

    def import_from_store(self, store_id: str) -> bool:
        files = self.window.core.api.google.store.import_store_files(store_id)
        for file in files:
            # Not used in this path
            pass
        return True

    def insert(self, store_id: str, data) -> RemoteFileItem:
        """
        Insert a Document into local DB

        :param store_id: store name ('fileSearchStores/...').
        :param data: document object from API
        """
        file = RemoteFileItem()
        # Use document name as unique id and file_id
        file.id = getattr(data, "name", None)
        file.file_id = getattr(data, "name", None)
        file.thread_id = ""
        file.name = getattr(data, "display_name", None) or getattr(data, "name", "")
        file.provider = self.PROVIDER_NAME
        file.path = getattr(data, "display_name", None) or getattr(data, "name", "")
        # sizeBytes is a string (int64), convert if possible
        try:
            file.size = int(getattr(data, "size_bytes", 0) or 0)
        except Exception:
            file.size = 0
        file.store_id = store_id
        file.record_id = self.provider.create(file)
        self.items[file.id] = file
        return file

    def load(self):
        self.items = self.provider.load_all(self.PROVIDER_NAME)
        self.sort_items()

    def sort_items(self):
        return

    def save(self):
        self.provider.save_all(self.items)

    def get_version(self) -> str:
        return self.provider.get_version()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.24 23:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction, QActionGroup


class Menu:
    def __init__(self, window=None):
        """
        Theme menu controller

        :param window: Window instance
        """
        self.window = window
        self.density_values = (-2, -1, 0, 1, 2)
        self.loaded = False
        self.syntax_loaded = False
        self.density_loaded = False

        self._style_group = None
        self._theme_group = None
        self._density_group = None
        self._syntax_group = None

    def _on_style_triggered(self, action):
        """
        Handle style change triggered by menu action

        :param action: QAction instance
        """
        self.window.controller.theme.toggle_style(action.data())

    def _on_theme_triggered(self, action):
        """
        Handle theme change triggered by menu action

        :param action: QAction instance
        """
        self.window.controller.theme.toggle_theme_by_menu(action.data())

    def _on_syntax_triggered(self, action):
        """
        Handle syntax highlight change triggered by menu action

        :param action: QAction instance
        """
        self.window.controller.theme.toggle_syntax(action.data(), update_menu=True)

    def _on_density_triggered(self, action):
        """
        Handle layout density change triggered by menu action

        :param action: QAction instance
        """
        self.window.controller.theme.toggle_option_by_menu('layout.density', action.data())

    def setup_list(self):
        """Setup menu list"""
        if self.loaded:
            return

        w = self.window
        menu = w.ui.menu
        common = w.controller.theme.common

        if self._style_group is None:
            self._style_group = QActionGroup(w)
            self._style_group.setExclusive(True)
            self._style_group.triggered.connect(self._on_style_triggered)

        styles = common.get_styles_list()
        menu_style_dict = menu['theme_style']
        menu_style = menu['theme.style']
        for style in styles:
            style_id = style.lower()
            title = style.replace('_', ' ').title()
            if title == "Chatgpt":
                title = "ChatGPT"
            elif title == "Chatgpt Wide":
                title = "ChatGPT (wide)"
            act = QAction(title, w, checkable=True)
            act.setData(style_id)
            menu_style_dict[style_id] = act
            self._style_group.addAction(act)
            menu_style.addAction(act)

        if self._theme_group is None:
            self._theme_group = QActionGroup(w)
            self._theme_group.setExclusive(True)
            self._theme_group.triggered.connect(self._on_theme_triggered)

        themes = common.get_themes_list()
        themes += common.get_custom_themes_list()
        themes.sort()

        menu_theme_dict = menu['theme']
        menu_dark = menu['theme.dark']
        menu_light = menu['theme.light']
        for theme in themes:
            name = common.translate(theme)
            act = QAction(name, w, checkable=True)
            act.setData(theme)
            menu_theme_dict[theme] = act
            self._theme_group.addAction(act)
            if theme.startswith('dark'):
                menu_dark.addAction(act)
            elif theme.startswith('light'):
                menu_light.addAction(act)

        self.loaded = True

    def setup_syntax(self):
        """Setup syntax menu"""
        w = self.window
        menu = w.ui.menu

        styles = w.controller.chat.render.web_renderer.body.highlight.get_styles()
        styles.sort()

        if self.syntax_loaded:
            existing = sorted(menu['theme_syntax'].keys())
            if existing == styles:
                return

        menu_syntax_dict = menu['theme_syntax']
        menu_syntax = menu['theme.syntax']

        for act in menu_syntax_dict.values():
            menu_syntax.removeAction(act)
            act.deleteLater()
        menu_syntax_dict.clear()

        if self._syntax_group is None:
            self._syntax_group = QActionGroup(w)
            self._syntax_group.setExclusive(True)
            self._syntax_group.triggered.connect(self._on_syntax_triggered)

        for style in styles:
            act = QAction(style, w, checkable=True)
            act.setData(style)
            menu_syntax_dict[style] = act
            self._syntax_group.addAction(act)
            menu_syntax.addAction(act)

        self.syntax_loaded = True

    def setup_density(self):
        """Setup menu list"""
        if self.density_loaded:
            return

        w = self.window
        menu = w.ui.menu

        if self._density_group is None:
            self._density_group = QActionGroup(w)
            self._density_group.setExclusive(True)
            self._density_group.triggered.connect(self._on_density_triggered)

        current_density = w.core.config.get('layout.density')
        menu_density_dict = menu['theme.layout.density']
        menu_density = menu['theme.density']

        for value in self.density_values:
            name = str(value)
            if value > 0:
                name = '+' + name
            act = QAction(name, w, checkable=True)
            act.setData(value)
            menu_density_dict[value] = act
            self._density_group.addAction(act)
            menu_density.addAction(act)
            if value == current_density:
                act.setChecked(True)

        self.density_loaded = True

    def update_density(self):
        """Update layout density menu"""
        current_density = self.window.core.config.get('layout.density')
        items = self.window.ui.menu['theme.layout.density']
        act = items.get(current_density)
        if act is not None:
            act.setChecked(True)
        else:
            for a in items.values():
                a.setChecked(False)

    def update_list(self):
        """Update theme list menu"""
        current_style = self.window.core.config.get('theme.style')
        style_items = self.window.ui.menu['theme_style']
        act = style_items.get(current_style)
        if act is not None:
            act.setChecked(True)
        else:
            for a in style_items.values():
                a.setChecked(False)

        current_theme = self.window.core.config.get('theme')
        theme_items = self.window.ui.menu['theme']
        act = theme_items.get(current_theme)
        if act is not None:
            act.setChecked(True)
        else:
            for a in theme_items.values():
                a.setChecked(False)

    def update_syntax(self):
        """Update code syntax highlight menu"""
        current = self.window.core.config.get('render.code_syntax')
        items = self.window.ui.menu['theme_syntax']
        act = items.get(current)
        if act is not None:
            act.setChecked(True)
        else:
            for a in items.values():
                a.setChecked(False)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.26 19:00:00                  #
# ================================================== #

from typing import Optional, List

from llama_index.core.tools import FunctionTool

from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.bridge.worker import BridgeSignals
from pygpt_net.core.events import KernelEvent
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans

from .base import BaseRunner

class Loop(BaseRunner):
    def __init__(self, window=None):
        """
        Agent runner

        :param window: Window instance
        """
        super(Loop, self).__init__(window)
        self.window = window

        self.next_instruction = ""  # evaluation instruction
        self.prev_score = -1  # evaluation score

    def run_once(
            self,
            input: str,
            tools: List[FunctionTool],
            model_name: Optional[str] = None
    ) -> str:
        """
        Run agent once (quick call to ReAct agent)

        :param input: input text
        :param tools: tools
        :param model_name: model name
        :return: text response
        """
        if self.is_stopped():
            return ""  # abort if stopped

        model = self.window.core.models.get(model_name)
        ctx = CtxItem()
        bridge_context = BridgeContext(
            ctx=ctx,
            history=[],
            model=model,
            prompt=self.prepare_input(input),
            stream=False,
        )
        extra = {
            "agent_provider": "react",  # use React workflow provider
            "agent_tools": tools,
        }
        response_ctx = self.window.core.agents.runner.call_once(
            context=bridge_context,
            extra=extra,
            signals=None,
        )
        if response_ctx:
            return str(response_ctx.output)
        else:
            return "No response from evaluator."

    def run_next(
            self,
            context: BridgeContext,
            extra: dict,
            signals: BridgeSignals
    ) -> bool:
        """
        Evaluate a response and run next step

        :param context: BridgeContext
        :param extra: extra data
        :param signals: BridgeSignals
        """
        if self.is_stopped():
            return True  # abort if stopped

        ctx = context.ctx
        self.send_response(ctx, signals, KernelEvent.APPEND_BEGIN)  # lock input, show stop btn

        history = context.history
        tools = self.window.core.agents.observer.evaluation.get_tools()
        mode = self.window.core.config.get('agent.llama.loop.mode', "score")

        prompt = ""
        if mode == "score":
            prompt = self.window.core.agents.observer.evaluation.get_prompt_score(history)
        elif mode == "complete":
            prompt = self.window.core.agents.observer.evaluation.get_prompt_complete(history)

        # evaluate
        self.set_busy(signals)
        self.next_instruction = ""  # reset
        self.prev_score = -1  # reset

        # select evaluation model
        eval_model = ctx.model
        custom_model = self.window.core.config.get('agent.llama.eval_model', None)
        if custom_model and custom_model != "_":
            eval_model = custom_model

        if self.is_verbose():
            print("[Evaluation] Prompt:", prompt)
            print("[Evaluation] Running with model:", eval_model)

        # run agent once
        self.run_once(prompt, tools, eval_model)  # tool will update evaluation
        return self.handle_evaluation(ctx, self.next_instruction, self.prev_score, signals)

    def handle_evaluation(
            self,
            ctx: CtxItem,
            instruction: str,
            score: int,
            signals: BridgeSignals
    ):
        """
        Handle evaluation

        :param ctx: CtxItem
        :param instruction: instruction
        :param score: score
        :param signals: BridgeSignals
        """
        if self.is_stopped():
            return True  # abort if stopped

        score = int(score)
        msg = "{score_label}: {score}%".format(
            score_label=trans('eval.score'),
            score=str(score)
        )
        self.set_status(signals, msg)

        if self.is_verbose():
            print("[Evaluation] Score:", score)

        if score < 0:
            self.send_response(ctx, signals, KernelEvent.APPEND_END)
            self.set_idle(signals)
            return True
        good_score = self.window.core.config.get("agent.llama.loop.score", 75)
        if self.is_verbose():
            print("[Evaluation] Score needed:", good_score)
        if score >= good_score != 0:
            msg = "{status_finished} {score_label}: {score}%".format(
                status_finished=trans('status.finished'),
                score_label=trans('eval.score'),
                score=str(score)
            )
            ctx.extra["agent_eval_finish"] = True
            if self.is_verbose():
                print("[Evaluation] Stopping. Finish with score:", score)
            self.send_response(ctx, signals, KernelEvent.APPEND_END, msg=msg)
            self.set_idle(signals)
            return True

        # print("Instruction: " + instruction, "Score: " + str(score))
        step_ctx = self.add_ctx(ctx)
        step_ctx.set_input(instruction)
        step_ctx.set_output("")
        step_ctx.results = [
            {
                "loop": {
                    "score": score,
                }
            }
        ]
        step_ctx.extra = {
            "agent_input": True,
            "agent_evaluate": True,
            "footer": "Score: " + str(score) + "%",
        }
        step_ctx.internal = False  # input

        self.set_busy(signals)
        self.send_response(step_ctx, signals, KernelEvent.APPEND_DATA)

        # call next run
        context = BridgeContext()
        context.ctx = step_ctx
        context.history = self.window.core.ctx.all(meta_id=ctx.meta.id)
        context.prompt = instruction  # use instruction as prompt
        preset = self.window.controller.presets.get_current()
        context.preset = preset
        extra = {
            "agent_idx": preset.idx,
            "agent_provider": preset.agent_provider,
        }
        if preset.agent_openai:
            extra["agent_provider"] = preset.agent_provider_openai
        if self.is_verbose():
            print("[Evaluation] Instruction:", instruction)
            print("[Evaluation] Running next step...")
        context.model = self.window.core.models.get(self.window.core.config.get('model'))
        return self.window.core.agents.runner.call(context, extra, signals)

    def is_verbose(self) -> bool:
        """
        Check if verbose mode is enabled

        :return: True if verbose mode is enabled
        """
        return self.window.core.config.get("agent.llama.verbose", False)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.23 07:00:00                  #
# ================================================== #

import os.path
import re
import subprocess
import json
import os
import platform
import time
from typing import Any, Dict, List, Optional, Tuple

from PySide6.QtGui import QGuiApplication
from PySide6.QtCore import QRect

from pygpt_net.item.ctx import CtxItem


class Runner:
    def __init__(self, plugin=None):
        """
        Cmd Runner

        :param plugin: plugin
        """
        self.plugin = plugin
        self.signals = None
        self._winapi = None  # lazy
        self._winapi_mod = None  # lazy

    def attach_signals(self, signals):
        """
        Attach signals

        :param signals: signals
        """
        self.signals = signals

    # -------------------------------
    # Common helpers / logging
    # -------------------------------
    def handle_result(self, stdout, stderr):
        """
        Handle result from subprocess

        :param stdout: stdout
        :param stderr: stderr
        :return: result
        """
        result = None
        if stdout:
            result = stdout.decode("utf-8", errors="replace")
            self.log("STDOUT: {}".format(result))
        if stderr:
            err = stderr.decode("utf-8", errors="replace")
            # Prefer stderr if non-empty
            result = err if err else result
            self.log("STDERR: {}".format(err))
        if result is None:
            result = "No result (STDOUT/STDERR empty)"
            self.log(result)
        return result

    def handle_result_docker(self, response) -> str:
        """
        Handle result from docker container

        :param response: response
        :return: result
        """
        result = None
        if response:
            try:
                result = response.decode('utf-8', errors="replace")
            except Exception:
                result = str(response)
        self.log(
            "Result: {}".format(result),
            sandbox=True,
        )
        return result

    def is_sandbox(self) -> bool:
        """
        Check if sandbox is enabled

        :return: True if sandbox is enabled
        """
        return self.plugin.get_option_value('sandbox_docker')

    def get_docker(self) -> Any:
        """
        Get docker client

        :return: docker client instance
        """
        import docker
        return docker.from_env()

    def get_volumes(self) -> dict:
        """
        Get docker volumes

        :return: docker volumes
        """
        path = self.plugin.window.core.config.get_user_dir('data')
        mapping = {}
        mapping[path] = {
            "bind": "/data",
            "mode": "rw",
        }
        return mapping

    def run_docker(self, cmd: str) -> bytes or None:
        """
        Run docker container with command and return response

        :param cmd: command to run
        :return: response
        """
        client = self.get_docker()
        mapping = self.get_volumes()
        try:
            response = self.plugin.docker.execute(cmd)
        except Exception as e:
            response = str(e).encode("utf-8")
        return response

    def sys_exec_host(self, ctx: CtxItem, item: dict, request: dict) -> dict:
        """
        Execute system command on host
        """
        msg = "Executing system command: {}".format(item["params"]['command'])
        self.log(msg)
        self.log("Running command: {}".format(item["params"]['command']))
        try:
            process = subprocess.Popen(
                item["params"]['command'],
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, stderr = process.communicate()
        except Exception as e:
            self.error(e)
            stdout = None
            stderr = str(e).encode("utf-8")
        result = self.handle_result(stdout, stderr)
        return {
            "request": request,
            "result": str(result),
            "context": "SYS OUTPUT:\n--------------------------------\n" + self.parse_result(result),
        }

    def sys_exec_sandbox(self, ctx: CtxItem, item: dict, request: dict) -> dict:
        """
        Execute system command in sandbox (docker)
        """
        msg = "Executing system command: {}".format(item["params"]['command'])
        self.log(msg, sandbox=True)
        self.log(
            "Running command: {}".format(item["params"]['command']),
            sandbox=True,
        )
        response = self.run_docker(item["params"]['command'])
        result = self.handle_result_docker(response)
        return {
            "request": request,
            "result": str(result),
            "context": "SYS OUTPUT:\n--------------------------------\n" + self.parse_result(result),
        }

    def parse_result(self, result):
        """
        Parse result

        :param result: result
        :return: parsed result
        """
        if result is None:
            return ""
        img_ext = ["png", "jpg", "jpeg", "gif", "bmp", "tiff"]
        s = str(result).strip()
        if any(s.lower().endswith('.' + ext) for ext in img_ext):
            path = self.prepare_path(s.replace("file://", ""), on_host=True)
            if os.path.isfile(path):
                return "![Image](file://{})".format(path)
        return str(result)

    def is_absolute_path(self, path: str) -> bool:
        """
        Check if path is absolute
        """
        return os.path.isabs(path)

    def prepare_path(self, path: str, on_host: bool = True) -> str:
        """
        Prepare path

        :param path: path to prepare
        :param on_host: is on host
        :return: prepared path
        """
        if not path:
            return path
        if self.is_absolute_path(path):
            return path
        else:
            if not self.is_sandbox() or on_host:
                return os.path.join(
                    self.plugin.window.core.config.get_user_dir('data'),
                    path,
                )
            else:
                return path

    def error(self, err: any):
        """
        Log error message
        """
        if self.signals is not None:
            self.signals.error.emit(err)

    def status(self, msg: str):
        """
        Send status message
        """
        if self.signals is not None:
            self.signals.status.emit(msg)

    def debug(self, msg: any):
        """
        Log debug message
        """
        if self.signals is not None:
            self.signals.debug.emit(msg)

    def log(self, msg, sandbox: bool = False):
        """
        Log message to console
        """
        prefix = ''
        if sandbox:
            prefix += '[DOCKER]'
        full_msg = prefix + ' ' + str(msg)

        if self.signals is not None:
            self.signals.log.emit(full_msg)

    # -------------------------------
    # WinAPI helpers
    # -------------------------------
    def _ensure_windows(self):
        """Ensure the platform is Windows and WinAPI enabled."""
        if platform.system() != "Windows":
            raise RuntimeError("Windows API is available on Microsoft Windows only.")
        if not self.plugin.get_option_value("winapi_enabled"):
            raise RuntimeError("WinAPI is disabled in plugin options.")

    def _load_winapi_module(self):
        self._ensure_windows()
        if self._winapi_mod is not None:
            return self._winapi_mod
        try:
            from . import winapi as _win
        except Exception as e:
            raise RuntimeError("WinAPI backend unavailable: {}".format(e))
        self._winapi_mod = _win
        return self._winapi_mod

    def _ensure_winapi(self) -> Any:
        """Get or create WinAPI helper"""
        self._ensure_windows()
        if self._winapi is None:
            _win = self._load_winapi_module()
            self._winapi = _win.WinAPI()
        return self._winapi

    def _to_json(self, data: Any) -> str:
        return json.dumps(data, ensure_ascii=False, indent=2)

    def _resolve_window(self,
                        hwnd: Optional[int] = None,
                        title: Optional[str] = None,
                        exact: bool = False,
                        visible_only: bool = True,
                        class_name: Optional[str] = None,
                        exe: Optional[str] = None,
                        pid: Optional[int] = None) -> Tuple[Optional[int], Optional[List[Dict]], Optional[str]]:
        """
        Resolve a window handle by conditions. Returns (hwnd, candidates, error)
        """
        w = self._ensure_winapi()
        if hwnd:
            if w.is_window(hwnd):
                return hwnd, None, None
            return None, None, f"Window handle not valid: {hwnd}"

        items = w.enum_windows(visible_only=visible_only)
        def matches(item):
            ok = True
            if title:
                low = title.lower()
                ok = ok and ((item["title"] == title) if exact else (low in item["title"].lower()))
            if class_name:
                ok = ok and (item["class_name"].lower() == class_name.lower())
            if exe:
                ex = (item["exe"] or "")
                ok = ok and (os.path.basename(ex).lower() == os.path.basename(exe).lower())
            if pid is not None:
                ok = ok and (item["pid"] == int(pid))
            return ok

        matched = [it for it in items if matches(it)]
        if len(matched) == 0:
            return None, [], "No window matched."
        if len(matched) > 1:
            return None, matched, "Ambiguous: multiple windows matched."
        return matched[0]["hwnd"], None, None

    # -------------------------------
    # WinAPI: window listing / info
    # -------------------------------
    def win_list(self,
                 filter_title: Optional[str] = None,
                 visible_only: bool = True,
                 limit: Optional[int] = None) -> Dict:
        w = self._ensure_winapi()
        items = w.enum_windows(visible_only=visible_only)
        if filter_title:
            low = filter_title.lower()
            items = [x for x in items if low in x["title"].lower()]
        if limit and limit > 0:
            items = items[:limit]
        return {"result": self._to_json(items), "context": f"Found: {len(items)} windows"}

    def win_find(self,
                 title: Optional[str] = None,
                 class_name: Optional[str] = None,
                 exe: Optional[str] = None,
                 pid: Optional[int] = None,
                 exact: bool = False,
                 visible_only: bool = True) -> Dict:
        w = self._ensure_winapi()
        items = w.enum_windows(visible_only=visible_only)

        def matches(it):
            ok = True
            if title:
                low = title.lower()
                ok = ok and ((it["title"] == title) if exact else (low in it["title"].lower()))
            if class_name:
                ok = ok and (it["class_name"].lower() == class_name.lower())
            if exe:
                ex = (it["exe"] or "")
                ok = ok and (os.path.basename(ex).lower() == os.path.basename(exe).lower())
            if pid is not None:
                ok = ok and (it["pid"] == int(pid))
            return ok

        out = [it for it in items if matches(it)]
        return {"result": self._to_json(out), "context": f"Matches: {len(out)}"}

    def win_children(self, hwnd: int) -> Dict:
        w = self._ensure_winapi()
        if not hwnd:
            return {"result": "Missing hwnd", "context": "Param 'hwnd' is required."}
        kids = w.enum_child_windows(hwnd)
        return {"result": self._to_json(kids), "context": f"Children: {len(kids)}"}

    def win_foreground(self) -> Dict:
        w = self._ensure_winapi()
        hwnd = w.get_foreground_window()
        if not hwnd:
            return {"result": "None", "context": "No foreground window."}
        info = w.get_window_info(hwnd)
        return {"result": self._to_json(info), "context": f"Foreground HWND: {hwnd}"}

    def win_rect(self, hwnd: Optional[int] = None, title: Optional[str] = None, exact: bool = False) -> Dict:
        w = self._ensure_winapi()
        handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
        if candidates is not None:
            return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
        if err:
            return {"result": err, "context": err}
        r = w.get_window_rect(handle)
        return {"result": self._to_json(r), "context": f"Rect: {r}"}

    def win_get_state(self, hwnd: Optional[int] = None, title: Optional[str] = None, exact: bool = False) -> Dict:
        w = self._ensure_winapi()
        handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
        if candidates is not None:
            return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
        if err:
            return {"result": err, "context": err}
        info = w.get_window_info(handle)
        return {"result": self._to_json(info), "context": "Window state"}

    # -------------------------------
    # WinAPI: window control
    # -------------------------------
    def win_focus(self, hwnd: Optional[int] = None, title: Optional[str] = None, exact: bool = False) -> Dict:
        w = self._ensure_winapi()
        handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=True)
        if candidates is not None:
            return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
        if err:
            return {"result": err, "context": err}
        ok, msg = w.bring_to_foreground(handle)
        return {"result": "OK" if ok else "FAILED", "context": msg}

    def win_move_resize(self,
                        hwnd: Optional[int] = None,
                        title: Optional[str] = None,
                        exact: bool = False,
                        x: Optional[int] = None,
                        y: Optional[int] = None,
                        width: Optional[int] = None,
                        height: Optional[int] = None) -> Dict:
        w = self._ensure_winapi()
        handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
        if candidates is not None:
            return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
        if err:
            return {"result": err, "context": err}
        ok, msg = w.move_resize(handle, x, y, width, height)
        return {"result": "OK" if ok else "FAILED", "context": msg}

    def win_minimize(self, hwnd: Optional[int] = None, title: Optional[str] = None, exact: bool = False) -> Dict:
        w = self._ensure_winapi()
        handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
        if candidates is not None:
            return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
        if err:
            return {"result": err, "context": err}
        ok, msg = w.show_window(handle, state="minimize")
        return {"result": "OK" if ok else "FAILED", "context": msg}

    def win_maximize(self, hwnd: Optional[int] = None, title: Optional[str] = None, exact: bool = False) -> Dict:
        w = self._ensure_winapi()
        handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
        if candidates is not None:
            return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
        if err:
            return {"result": err, "context": err}
        ok, msg = w.show_window(handle, state="maximize")
        return {"result": "OK" if ok else "FAILED", "context": msg}

    def win_restore(self, hwnd: Optional[int] = None, title: Optional[str] = None, exact: bool = False) -> Dict:
        w = self._ensure_winapi()
        handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
        if candidates is not None:
            return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
        if err:
            return {"result": err, "context": err}
        ok, msg = w.show_window(handle, state="restore")
        return {"result": "OK" if ok else "FAILED", "context": msg}

    def win_close(self, hwnd: Optional[int] = None, title: Optional[str] = None, exact: bool = False) -> Dict:
        w = self._ensure_winapi()
        handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
        if candidates is not None:
            return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
        if err:
            return {"result": err, "context": err}
        ok, msg = w.close_window(handle)
        return {"result": "OK" if ok else "FAILED", "context": msg}

    def win_show(self, hwnd: Optional[int] = None, title: Optional[str] = None, exact: bool = False) -> Dict:
        w = self._ensure_winapi()
        handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
        if candidates is not None:
            return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
        if err:
            return {"result": err, "context": err}
        ok, msg = w.show_window(handle, state="show")
        return {"result": "OK" if ok else "FAILED", "context": msg}

    def win_hide(self, hwnd: Optional[int] = None, title: Optional[str] = None, exact: bool = False) -> Dict:
        w = self._ensure_winapi()
        handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
        if candidates is not None:
            return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
        if err:
            return {"result": err, "context": err}
        ok, msg = w.show_window(handle, state="hide")
        return {"result": "OK" if ok else "FAILED", "context": msg}

    def win_always_on_top(self,
                          topmost: bool,
                          hwnd: Optional[int] = None,
                          title: Optional[str] = None,
                          exact: bool = False) -> Dict:
        w = self._ensure_winapi()
        if topmost is None:
            return {"result": "Missing 'topmost'", "context": "Param 'topmost' is required (true/false)."}
        handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
        if candidates is not None:
            return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
        if err:
            return {"result": err, "context": err}
        ok, msg = w.set_topmost(handle, bool(topmost))
        return {"result": "OK" if ok else "FAILED", "context": msg}

    def win_set_opacity(self,
                        alpha: Optional[int] = None,
                        opacity: Optional[float] = None,
                        hwnd: Optional[int] = None,
                        title: Optional[str] = None,
                        exact: bool = False) -> Dict:
        w = self._ensure_winapi()
        handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
        if candidates is not None:
            return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
        if err:
            return {"result": err, "context": err}
        if alpha is None:
            if opacity is None:
                return {"result": "Missing 'alpha' or 'opacity'", "context": "Provide alpha (0..255) or opacity (0..1)."}
            alpha = int(max(0, min(1.0, float(opacity))) * 255.0)
        else:
            alpha = int(max(0, min(255, int(alpha))))
        ok, msg = w.set_opacity(handle, alpha)
        return {"result": "OK" if ok else "FAILED", "context": msg}

    # -------------------------------
    # WinAPI: screenshots
    # -------------------------------
    def _save_pixmap(self, pix, path: str) -> Tuple[bool, str]:
        """Save QPixmap to disk, ensure dir."""
        abspath = self.prepare_path(path)
        try:
            os.makedirs(os.path.dirname(abspath), exist_ok=True)
        except Exception:
            pass
        ok = pix.save(abspath, "PNG")
        return ok, abspath

    def win_screenshot(self,
                       hwnd: Optional[int] = None,
                       title: Optional[str] = None,
                       exact: bool = False,
                       path: Optional[str] = None) -> Dict:
        self._ensure_windows()
        handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
        if candidates is not None:
            return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
        if err:
            return {"result": err, "context": err}

        if not path:
            ts = time.strftime("%Y%m%d_%H%M%S")
            path = f"win_screenshot_{handle}_{ts}.png"

        screen = QGuiApplication.primaryScreen()
        if screen is None:
            return {"result": "No screen", "context": "No QScreen available. Is Qt app running?"}

        pix = screen.grabWindow(int(handle))
        if pix.isNull():
            return {"result": "Failed", "context": "grabWindow returned null pixmap."}

        ok, abspath = self._save_pixmap(pix, path)
        if not ok:
            return {"result": "Failed", "context": f"Could not save screenshot to: {abspath}"}
        context = "SYS OUTPUT:\n--------------------------------\n" + self.parse_result(abspath)
        return {"result": abspath, "context": context}

    def win_area_screenshot(self,
                            x: int, y: int, width: int, height: int,
                            hwnd: Optional[int] = None,
                            title: Optional[str] = None,
                            exact: bool = False,
                            relative: bool = False,
                            path: Optional[str] = None) -> Dict:
        self._ensure_windows()
        if any(v is None for v in [x, y, width, height]):
            return {"result": "Missing geometry", "context": "Params x,y,width,height are required."}

        x0, y0 = int(x), int(y)
        if relative and (hwnd or title):
            handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
            if candidates is not None:
                return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
            if err:
                return {"result": err, "context": err}
            rect = self._ensure_winapi().get_window_rect(handle)
            x0 += int(rect["left"])
            y0 += int(rect["top"])

        if not path:
            ts = time.strftime("%Y%m%d_%H%M%S")
            path = f"win_area_{x0}_{y0}_{width}x{height}_{ts}.png"

        screen = QGuiApplication.primaryScreen()
        if screen is None:
            return {"result": "No screen", "context": "No QScreen available. Is Qt app running?"}

        pix = screen.grabWindow(0, x0, y0, int(width), int(height))
        if pix.isNull():
            return {"result": "Failed", "context": "grabWindow returned null pixmap."}

        ok, abspath = self._save_pixmap(pix, path)
        if not ok:
            return {"result": "Failed", "context": f"Could not save screenshot to: {abspath}"}
        context = "SYS OUTPUT:\n--------------------------------\n" + self.parse_result(abspath)
        return {"result": abspath, "context": context}

    # -------------------------------
    # WinAPI: clipboard / cursor / input / monitors
    # -------------------------------
    def win_clipboard_get(self) -> Dict:
        cb = QGuiApplication.clipboard()
        text = cb.text()
        return {"result": text, "context": "Clipboard text retrieved."}

    def win_clipboard_set(self, text: str) -> Dict:
        cb = QGuiApplication.clipboard()
        cb.setText(text or "")
        return {"result": "OK", "context": "Clipboard text set."}

    def win_cursor_get(self) -> Dict:
        w = self._ensure_winapi()
        x, y = w.get_cursor_pos()
        return {"result": self._to_json({"x": x, "y": y}), "context": f"Cursor: ({x}, {y})"}

    def win_cursor_set(self, x: int, y: int) -> Dict:
        w = self._ensure_winapi()
        ok = w.set_cursor_pos(x, y)
        return {"result": "OK" if ok else "FAILED", "context": f"Set cursor to ({x}, {y})"}

    def win_keys_text(self, text: str, per_char_delay_ms: Optional[int] = None) -> Dict:
        self._ensure_windows()
        if not text:
            return {"result": "No text", "context": "Param 'text' is required."}
        if per_char_delay_ms is None:
            per_char_delay_ms = int(self.plugin.get_option_value("win_keys_per_char_delay_ms"))
        _win = self._load_winapi_module()
        sender = _win.InputSender()
        sender.send_unicode_text(text, per_char_delay_ms=per_char_delay_ms)
        return {"result": "OK", "context": f"Typed {len(text)} characters."}

    def win_keys_send(self,
                      keys: List[str],
                      hold_ms: Optional[int] = None,
                      gap_ms: Optional[int] = None) -> Dict:
        self._ensure_windows()
        if not keys or not isinstance(keys, list):
            return {"result": "No keys", "context": "Param 'keys' must be a non-empty list of key tokens."}
        if hold_ms is None:
            hold_ms = int(self.plugin.get_option_value("win_keys_hold_ms"))
        if gap_ms is None:
            gap_ms = int(self.plugin.get_option_value("win_keys_gap_ms"))
        _win = self._load_winapi_module()
        sender = _win.InputSender()
        ok, msg = sender.send_keys(keys, hold_ms=hold_ms, gap_ms=gap_ms)
        return {"result": "OK" if ok else "FAILED", "context": msg}

    def win_click(self,
                  x: Optional[int] = None,
                  y: Optional[int] = None,
                  button: str = "left",
                  double: bool = False,
                  hwnd: Optional[int] = None,
                  title: Optional[str] = None,
                  exact: bool = False,
                  relative: bool = False) -> Dict:
        self._ensure_windows()
        if x is None or y is None:
            return {"result": "Missing coords", "context": "Params 'x' and 'y' are required."}
        x0, y0 = int(x), int(y)
        if relative and (hwnd or title):
            handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
            if candidates is not None:
                return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
            if err:
                return {"result": err, "context": err}
            rect = self._ensure_winapi().get_window_rect(handle)
            x0 += int(rect["left"])
            y0 += int(rect["top"])
        _win = self._load_winapi_module()
        sender = _win.InputSender()
        ok, msg = sender.click_at(x0, y0, button=button, double=bool(double))
        return {"result": "OK" if ok else "FAILED", "context": msg}

    def win_drag(self,
                 x1: int, y1: int, x2: int, y2: int,
                 hwnd: Optional[int] = None,
                 title: Optional[str] = None,
                 exact: bool = False,
                 relative: bool = False,
                 steps: int = 20,
                 hold_ms: Optional[int] = None) -> Dict:
        self._ensure_windows()
        if any(v is None for v in [x1, y1, x2, y2]):
            return {"result": "Missing coords", "context": "Params x1,y1,x2,y2 are required."}
        sx, sy, ex, ey = int(x1), int(y1), int(x2), int(y2)
        if relative and (hwnd or title):
            handle, candidates, err = self._resolve_window(hwnd, title, exact=exact, visible_only=False)
            if candidates is not None:
                return {"result": self._to_json(candidates), "context": "Multiple matches. Specify 'hwnd'."}
            if err:
                return {"result": err, "context": err}
            rect = self._ensure_winapi().get_window_rect(handle)
            offx, offy = int(rect["left"]), int(rect["top"])
            sx, sy, ex, ey = sx + offx, sy + offy, ex + offx, ey + offy
        if hold_ms is None:
            hold_ms = int(self.plugin.get_option_value("win_drag_step_delay_ms"))
        _win = self._load_winapi_module()
        sender = _win.InputSender()
        ok, msg = sender.drag_and_drop(sx, sy, ex, ey, steps=max(1, int(steps)), step_delay_ms=int(hold_ms))
        return {"result": "OK" if ok else "FAILED", "context": msg}

    def win_monitors(self) -> Dict:
        screens = QGuiApplication.screens()
        arr = []
        for i, s in enumerate(screens):
            g: QRect = s.geometry()
            arr.append({
                "index": i,
                "name": s.name(),
                "geometry": {"x": g.x(), "y": g.y(), "width": g.width(), "height": g.height()},
                "dpi": s.logicalDotsPerInch(),
                "scale": s.devicePixelRatio(),
            })
        return {"result": self._to_json(arr), "context": f"Monitors: {len(arr)}"}


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.25 03:00:00                  #
# ================================================== #

import os

from PySide6.QtCore import Qt, QUrl, QTimer
from PySide6.QtGui import QIcon, QAction
from PySide6.QtWidgets import (QWidget, QPushButton, QHBoxLayout, QVBoxLayout,
                               QFileDialog, QSlider, QLabel, QStyle, QSizePolicy, QMenu, QMessageBox)

from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.utils import trans

class VideoPlayerWidget(QWidget):
    def __init__(self, window=None):
        super().__init__(window)
        self.window = window
        self.player = None
        self.audio = None
        self.video = None
        self.loaded = False
        self.path = None
        self.stopped = False
        self.seeking = False

        self._QMediaPlayer = None
        self._QAudioOutput = None
        self._QVideoWidget = None

        self._waiting_position_set = False
        self._last_position = -1

        self.update_timer = QTimer(self)
        self.update_timer.setInterval(200)
        self.update_timer.timeout.connect(self.update_ui)
        self.autoplay_timer = QTimer(self)
        self.autoplay_timer.setInterval(1000)
        self.autoplay_timer.setSingleShot(True)
        self.autoplay_timer.timeout.connect(self.after_loaded)

        self._video_placeholder = QWidget(window)
        self._video_placeholder.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self._video_placeholder.setMinimumSize(640, 480)

        self._icon_play = self.style().standardIcon(QStyle.SP_MediaPlay)
        self._icon_pause = self.style().standardIcon(QStyle.SP_MediaPause)
        self._icon_stop = self.style().standardIcon(QStyle.SP_MediaStop)

        self.btn_play_pause = QPushButton()
        self.btn_play_pause.setEnabled(False)

        self.btn_stop = QPushButton()
        self.btn_stop.setIcon(self._icon_stop)
        self.btn_stop.setEnabled(False)

        self.btn_mute = QPushButton()
        self.btn_mute.setIcon(QIcon(':/icons/mute.svg'))
        self.btn_mute.setCheckable(True)

        self.btn_open = QPushButton(trans("action.video.open"))

        self.label_duration = QLabel()
        self.label_time = QLabel()
        self.label_path = HelpLabel("")
        self.label_path.setWordWrap(False)

        self.slider = QSlider(Qt.Horizontal)
        self.slider.setRange(0, 0)

        self.volume_slider = QSlider(Qt.Horizontal)
        self.volume_slider.setRange(0, 100)
        self.volume_slider.setValue(100)
        self.volume_slider.setToolTip('Adjust volume')

        layout_volume = QHBoxLayout()
        layout_volume.addWidget(self.volume_slider)
        layout_volume.addWidget(self.btn_mute)

        layout_bottom = QHBoxLayout()
        layout_bottom.addWidget(self.label_path)
        layout_bottom.addStretch()
        layout_bottom.addLayout(layout_volume)

        layout_controls = QHBoxLayout()
        layout_controls.addWidget(self.btn_open)
        layout_controls.addWidget(self.btn_play_pause)
        layout_controls.addWidget(self.btn_stop)

        layout_slider = QHBoxLayout()
        layout_slider.addWidget(self.label_time)
        layout_slider.addWidget(self.slider)
        layout_slider.addWidget(self.label_duration)

        layout = QVBoxLayout()
        layout.addWidget(self._video_placeholder)
        layout.addLayout(layout_slider)
        layout.addLayout(layout_controls)
        layout.addLayout(layout_bottom)
        self._layout_main = layout

        self.setLayout(layout)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.on_context_menu)

        self.update_play_pause_icon()
        self.bind_signals()

    def _ensure_multimedia(self):
        """Ensure multimedia components are loaded"""
        if self.player is not None:
            return
        from PySide6.QtMultimedia import QMediaPlayer, QAudioOutput
        from PySide6.QtMultimediaWidgets import QVideoWidget
        self._QMediaPlayer = QMediaPlayer
        self._QAudioOutput = QAudioOutput
        self._QVideoWidget = QVideoWidget
        self.player = QMediaPlayer(self.window)
        self.audio = QAudioOutput()
        self.player.setAudioOutput(self.audio)
        self.video = QVideoWidget(self.window)
        self.video.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.video.setMinimumSize(640, 480)
        idx = self._layout_main.indexOf(self._video_placeholder)
        if idx != -1:
            self._layout_main.removeWidget(self._video_placeholder)
            self._video_placeholder.setParent(None)
            self._layout_main.insertWidget(idx, self.video)
        else:
            self._layout_main.insertWidget(0, self.video)
        self.video.mousePressEvent = self.video_widget_clicked
        self.player.setVideoOutput(self.video)
        self.player.positionChanged.connect(self.position_changed)
        self.player.durationChanged.connect(self.duration_changed)
        self.player.playbackStateChanged.connect(self.on_playback_state_changed)
        self.player.mediaStatusChanged.connect(self.media_status_changed)
        self.update_mute_icon()
        self.update_volume_slider()
        self.adjust_volume(self.volume_slider.value())
        self.set_muted(self.btn_mute.isChecked())

    def bind_signals(self):
        """Bind signals to player controls"""
        self.btn_open.clicked.connect(self.open_file)
        self.btn_play_pause.clicked.connect(self.toggle_play_pause)
        self.btn_stop.clicked.connect(self.stop_video)
        self.btn_mute.toggled.connect(self.set_muted)
        self.volume_slider.valueChanged.connect(self.adjust_volume)
        self.slider.sliderPressed.connect(self.on_slider_pressed)
        self.slider.sliderReleased.connect(self.on_slider_released)
        self.slider.valueChanged.connect(self.on_slider_value_changed)
        self._video_placeholder.mousePressEvent = self.video_widget_clicked
        self.seeking = False

    def update(self):
        """Update player state"""
        self.update_mute_icon()
        self.update_volume_slider()
        self.update_label_path()

    def open_file(self):
        """Open video file dialog"""
        path, _ = QFileDialog.getOpenFileName(
            self,
            trans("action.video.open"),
            "",
            "Video and Audio Files (*.mp4 *.avi *.mkv *.mp3 *.wav *.flac *.ogg *.m4a *.aac *.wma *.webm *.mkv "
            "*.flv *.mov *.wmv *.3gp *.m4v *.mpg);;All Files (*)"
        )
        if path:
            self.open(path)

    def open(self, path: str):
        """
        Open video

        :param path: video path
        """
        self.loaded = False
        self.stopped = False
        self.seeking = False
        self.reset()
        self._ensure_multimedia()
        self.player.setSource(QUrl.fromLocalFile(path))
        self.set_path(path)
        self.enable()
        self.autoplay_timer.start()
        self.update()
        self.window.tools.get("player").store_path(path)

    def enable(self):
        """Enable player"""
        self.btn_play_pause.setEnabled(True)
        self.btn_stop.setEnabled(True)

    def play_video(self):
        """Play video"""
        self.stopped = False
        if not self.loaded and self.path:
            self.open(self.path)
        else:
            self._ensure_multimedia()
            self.update_audio()
            self.player.play()

    def stop_video(self):
        """Stop video"""
        if self.player and self.player.source():
            self.player.stop()
        self.stopped = True
        self.slider.setValue(0)
        self.label_time.setText(self.format_time(0))
        self.btn_play_pause.setIcon(self._icon_play)
        self.update_timer.stop()

    def toggle_play_pause(self):
        """Toggle play/pause"""
        if self.player and self._QMediaPlayer and self.player.playbackState() == self._QMediaPlayer.PlayingState:
            self.player.pause()
        else:
            self.play_video()
        self.on_playback_state_changed()

    def reset(self):
        """Reset player to default state"""
        if self.player and self.player.source():
            self.player.stop()
            self.player.setSource(QUrl())
        self.btn_play_pause.setIcon(self._icon_play)
        self.update_timer.stop()
        self.slider.setValue(0)
        self.seeking = False
        self.loaded = False
        self._waiting_position_set = False
        self._last_position = -1

    def reset_player(self):
        """Reset player to default state"""
        if self.player is not None:
            if self._QMediaPlayer and self.player.playbackState() == self._QMediaPlayer.PlayingState:
                self.player.stop()
            self.player.setSource(QUrl())

    def on_context_menu(self, point):
        """
        Context menu event

        :param point: point
        """
        if not self.path or not (self.player and self.player.source()):
            return
        context_menu = QMenu(self)
        save_as_action = QAction(QIcon(":/icons/save.svg"), trans("action.save_as"), self)
        save_as_action.triggered.connect(
            lambda: self.window.tools.get("player").save_as_file()
        )
        context_menu.addAction(save_as_action)
        context_menu.exec(self.mapToGlobal(point))

    def use_as_attachment(self):
        """Use as attachment"""
        path = self.window.tools.get("player").grab_frame()
        self.window.controller.files.use_attachment(path)

    def use_as_image(self):
        """Use as image"""
        path = self.window.tools.get("player").grab_frame()
        self.window.controller.painter.open_external(path)

    def adjust_volume(self, value):
        """
        Adjust volume

        :param value: volume value
        """
        self.window.core.config.set("video.player.volume", value)
        if self.audio:
            volume = value / 100.0
            self.audio.setVolume(volume)

    def set_path(self, path: str):
        """
        Set path to video

        :param path: path
        """
        self.path = path
        self.update_label_path()
        if not self.loaded and self.path:
            if not self.player or not self.player.source():
                self._ensure_multimedia()
                self.reset()
                self.player.setSource(QUrl.fromLocalFile(self.path))
            self.enable()
            self.update()
        self.window.ui.dialog["video_player"].setWindowTitle(os.path.basename(self.path))

    def on_close(self):
        """Stop video"""
        self.reset()
        self.autoplay_timer.stop()

    def toggle_fullscreen(self):
        """Toggle fullscreen"""
        if self.window.ui.dialog['video_player'].isFullScreen():
            self.window.ui.dialog['video_player'].showNormal()
        else:
            self.window.ui.dialog['video_player'].showFullScreen()

    def media_status_changed(self, status):
        """
        Media status changed

        :param status: status
        """
        if not self._QMediaPlayer:
            return
        if status == self._QMediaPlayer.LoadedMedia:
            self.loaded = True
            self.force_resize()
        elif status == self._QMediaPlayer.InvalidMedia:
            self.loaded = False
            self.window.ui.dialogs.alert("Failed to load media file, missing video codec?")

    def after_loaded(self):
        """Start playback after loaded"""
        if self.player and self._QMediaPlayer and self.player.mediaStatus() == self._QMediaPlayer.LoadedMedia:
            self.loaded = True
            if self.player.playbackState() != self._QMediaPlayer.PlayingState:
                self.toggle_play_pause()
                self.force_resize()

    def force_resize(self):
        """Force resize fix"""
        if not self.video:
            return
        self.video.hide()
        self.video.show()
        self.video.update()
        current_size = self.window.ui.dialog['video_player'].size()
        self.window.ui.dialog['video_player'].resize(current_size.width() + 10, current_size.height() + 10)
        self.window.ui.dialog['video_player'].resize(current_size)
        self.window.ui.dialog['video_player'].repaint()
        w = self.video.width() - 1
        h = self.video.height() - 1
        self.video.resize(w, h)

    def on_playback_state_changed(self):
        """Playback state changed"""
        self.update_play_pause_icon()

    def video_widget_clicked(self, event):
        """
        Video widget clicked

        :param event: event
        """
        if event.button() == Qt.LeftButton:
            if self.loaded:
                self.toggle_play_pause()
        elif event.button() == Qt.RightButton:
            pass

    def duration_changed(self, duration):
        """
        Duration changed

        :param duration: duration
        """
        self.slider.setRange(0, duration)
        self.label_duration.setText(self.format_time(duration))

    def set_muted(self, state: bool):
        """
        Set volume muted

        :param state: muted
        """
        if self.audio:
            self.audio.setMuted(state)
        self.window.core.config.set("video.player.volume.mute", state)

    def set_position(self, position):
        """
        Set position

        :param position: position
        """
        if not self.player:
            return
        if self._QMediaPlayer and self.player.playbackState() == self._QMediaPlayer.PlayingState:
            self.player.pause()
        self._waiting_position_set = True
        self.player.setPosition(position)

    def on_slider_pressed(self):
        """Slider pressed"""
        self.seeking = True
        self._waiting_position_set = False
        if self.player:
            self.player.pause()

    def on_slider_value_changed(self, value):
        """Slider value changed"""
        if self.seeking:
            self.label_time.setText(self.format_time(value))

    def on_slider_released(self):
        """Slider released"""
        if self.seeking:
            if self.player:
                self.player.setPosition(self.slider.value())
                self.update_audio()
                QTimer.singleShot(0, self.player.play)
            self.seeking = False

    def position_changed(self, position):
        """
        Position changed

        :param position: position
        """
        if self._waiting_position_set and self.player:
            self._waiting_position_set = False
            if self._QMediaPlayer and self.player.playbackState() == self._QMediaPlayer.PausedState:
                QTimer.singleShot(100, self.player.play)
        if not self.seeking:
            self._last_position = position
            self.slider.setValue(position)
            self.label_time.setText(self.format_time(position))

    def update_label_path(self):
        """Update label path"""
        if self.path:
            self.label_path.setText(os.path.basename(self.path))
        else:
            self.label_path.setText("")

    def update_play_pause_icon(self):
        """Update play/pause icon"""
        if self.player and self._QMediaPlayer and self.player.playbackState() == self._QMediaPlayer.PlayingState:
            self.btn_play_pause.setIcon(self._icon_pause)
            if not self.update_timer.isActive():
                self.update_timer.start()
        else:
            self.btn_play_pause.setIcon(self._icon_play)
            if self.update_timer.isActive():
                self.update_timer.stop()

    def update_audio(self):
        """Re-assign audio"""
        if self.player and self.audio:
            try:
                current = self.player.audioOutput()
            except Exception:
                current = None
            if current is not self.audio:
                self.player.setAudioOutput(self.audio)

    def update_mute_icon(self):
        """Update mute icon"""
        if self.audio:
            muted = self.audio.isMuted()
            if self.btn_mute.isChecked() != muted:
                block = self.btn_mute.blockSignals(True)
                self.btn_mute.setChecked(muted)
                self.btn_mute.blockSignals(block)

    def update_volume_slider(self):
        """Update volume slider"""
        if self.audio:
            v = int(self.audio.volume() * 100)
            if self.volume_slider.value() != v:
                block = self.volume_slider.blockSignals(True)
                self.volume_slider.setValue(v)
                self.volume_slider.blockSignals(block)

    def update_ui(self):
        """Update UI"""
        if self.player and not self.seeking:
            pos = self.player.position()
            if pos != self._last_position:
                self._last_position = pos
                self.slider.setValue(pos)
                self.label_time.setText(self.format_time(pos))

    def format_time(self, ms):
        """
        Format time

        :param ms: milliseconds
        """
        seconds = (ms // 1000) % 60
        minutes = (ms // 60000) % 60
        hours = (ms // 3600000)
        return "{:02d}:{:02d}:{:02d}".format(hours, minutes, seconds)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.02 20:00:00                  #
# ================================================== #

import os

from PySide6 import QtCore
from PySide6.QtGui import QStandardItemModel, Qt, QIcon
from PySide6.QtWidgets import QVBoxLayout, QPushButton, QHBoxLayout, QWidget

from pygpt_net.ui.widget.element.button import SyncButton
from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.ui.widget.lists.uploaded import UploadedFileList
from pygpt_net.utils import trans

class AttachmentsUploaded:
    def __init__(self, window=None):
        """
        Attachments Uploaded UI

        :param window: Window instance
        """
        self.window = window
        self.id = 'attachments_uploaded'

    def setup(self) -> QVBoxLayout:
        """
        Setup list

        :return: QVBoxLayout
        """
        self.setup_attachments()
        empty_widget = QWidget(self.window)

        self.window.ui.nodes['attachments_uploaded.sync.tip'] = HelpLabel(trans('attachments_uploaded.sync.tip'), self.window)
        self.window.ui.nodes['attachments_uploaded.sync.tip'].setWordWrap(False)
        self.window.ui.nodes['attachments_uploaded.sync.tip'].setAlignment(Qt.AlignCenter)

        self.window.ui.nodes['tip.input.attachments.uploaded'] = HelpLabel(trans('tip.input.attachments.uploaded'),
                                                                           self.window)

        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(self.window.ui.nodes['attachments_uploaded.btn.sync'])
        buttons_layout.addWidget(self.window.ui.nodes['attachments_uploaded.btn.clear'])
        buttons_layout.addWidget(empty_widget)
        buttons_layout.addWidget(self.window.ui.nodes['attachments_uploaded.sync.tip'])
        buttons_layout.addStretch()

        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.nodes['tip.input.attachments.uploaded'])
        layout.addWidget(self.window.ui.nodes['attachments_uploaded'])
        layout.addLayout(buttons_layout)

        return layout

    def setup_attachments(self):
        """
        Setup attachments uploaded list
        """
        self.window.ui.nodes[self.id] = UploadedFileList(self.window)

        self.window.ui.nodes['attachments_uploaded.btn.sync'] = SyncButton(trans('attachments_uploaded.btn.sync'), self.window)
        self.window.ui.nodes['attachments_uploaded.btn.clear'] = QPushButton(QIcon(":/icons/close.svg"), trans('attachments_uploaded.btn.clear'))
        self.window.ui.nodes['attachments_uploaded.btn.clear'].clicked.connect(
            lambda: self.window.controller.assistant.files.clear()
        )

        self.window.ui.models[self.id] = self.create_model(self.window)
        self.window.ui.nodes[self.id].setModel(self.window.ui.models[self.id])

    def create_model(self, parent) -> QStandardItemModel:
        """
        Create list model

        :param parent: parent widget
        :return: QStandardItemModel
        """
        model = QStandardItemModel(0, 4, parent)
        model.setHeaderData(0, Qt.Horizontal, trans('attachments.header.name'))
        model.setHeaderData(1, Qt.Horizontal, trans('attachments.header.path'))
        model.setHeaderData(2, Qt.Horizontal, trans('attachments.header.size'))
        model.setHeaderData(3, Qt.Horizontal, trans('attachments.header.store'))
        return model

    def update(self, data):
        """
        Update list

        :param data: Data to update
        """
        model = self.window.ui.models[self.id]
        store_names = self.window.core.remote_store.openai.get_names()
        thread_only_label = trans("remote_store.thread_only")
        fs = self.window.core.filesystem

        model.beginResetModel()
        model.setRowCount(0)
        count = len(data)
        if count:
            model.setRowCount(count)
            for i, item in enumerate(data.values()):
                path = item.path
                if item.size is not None:
                    size_str = fs.sizeof_fmt(item.size)
                else:
                    size_str = "-"
                    if path:
                        try:
                            st = os.stat(path)
                        except OSError:
                            pass
                        else:
                            size_str = fs.sizeof_fmt(st.st_size)

                store_name = store_names.get(item.store_id, thread_only_label)

                idx0 = model.index(i, 0)
                model.setData(idx0, item.name)
                model.setData(idx0, f"file_id: {item.file_id}", QtCore.Qt.ToolTipRole)
                model.setData(model.index(i, 1), path)
                model.setData(model.index(i, 2), size_str)
                model.setData(model.index(i, 3), store_name)
        model.endResetModel()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.25 11:30:00                  #
# ================================================== #

from __future__ import annotations

try:
    from pygpt_net.utils import trans as _trans
except Exception:
    _trans = None


class EditorConfig:
    """Central place for all user-facing strings used by NodeEditor.

    Each method returns a final string. If a translator (pygpt_net.utils.trans)
    is available, keys are looked up; otherwise English defaults are used.
    """

    def _t(self, key: str, default: str) -> str:
        try:
            if _trans is None:
                return default
            val = _trans(key)
            # Fallback to default when translator returns key or empty
            if not val or val == key:
                return default
            return val
        except Exception:
            return default

    # ----- Commands / Undo text -----

    def cmd_add_node(self, type_name: str) -> str:
        return f"{self._t('node.editor.cmd.add', 'Add')} {type_name}"

    def cmd_move_node(self) -> str:
        return self._t('node.editor.cmd.move_node', 'Move Node')

    def cmd_resize_node(self) -> str:
        return self._t('node.editor.cmd.resize_node', 'Resize Node')

    def cmd_connect(self) -> str:
        return self._t('node.editor.cmd.connect', 'Connect')

    def cmd_rewire_connection(self) -> str:
        return self._t('node.editor.cmd.rewire_connection', 'Rewire Connection')

    def cmd_delete_connection(self) -> str:
        return self._t('node.editor.cmd.delete_connection', 'Delete Connection')

    def cmd_delete_node(self) -> str:
        return self._t('node.editor.cmd.delete_node', 'Delete Node')

    def cmd_clear(self) -> str:
        return self._t('node.editor.cmd.clear', 'Clear')

    def macro_delete_selection(self) -> str:
        return self._t('node.editor.macro.delete_selection', 'Delete selection')

    # ----- Scene / Context menus -----

    def menu_add(self) -> str:
        return self._t('node.editor.menu.add', 'Add')

    def menu_undo(self) -> str:
        return self._t('node.editor.menu.undo', 'Undo')

    def menu_redo(self) -> str:
        return self._t('node.editor.menu.redo', 'Redo')

    def menu_clear(self) -> str:
        return self._t('node.editor.menu.clear', 'Clear')

    def edge_context_delete(self) -> str:
        return self._t('node.editor.edge.delete', 'Delete connection')

    def node_context_rename(self) -> str:
        return self._t('node.editor.node.rename', 'Rename')

    def node_context_delete(self) -> str:
        return self._t('node.editor.node.delete', 'Delete')

    def rename_dialog_title(self) -> str:
        return self._t('node.editor.rename.title', 'Rename Node')

    def rename_dialog_label(self) -> str:
        return self._t('node.editor.rename.label', 'Name:')

    # ----- Overlays / Status -----

    def overlay_grab_tooltip(self) -> str:
        return self._t('node.editor.overlay.grab', 'Grab: toggle global panning with Left Mouse Button')

    def overlay_zoom_in_tooltip(self) -> str:
        return self._t('node.editor.overlay.zoom_in', 'Zoom In')

    def overlay_zoom_out_tooltip(self) -> str:
        return self._t('node.editor.overlay.zoom_out', 'Zoom Out')

    def status_no_nodes(self) -> str:
        return self._t('node.editor.status.no_nodes', 'No nodes')

    def type_unknown(self) -> str:
        return self._t('node.editor.type.unknown', 'Unknown')

    # ----- Labels / Tooltips -----

    def label_id(self) -> str:
        return self._t('node.editor.label.id', 'ID')

    def lbl_node(self) -> str:
        return self._t('node.editor.lbl.node', 'Node:')

    def lbl_port(self) -> str:
        return self._t('node.editor.lbl.port', 'Port:')

    def lbl_allowed_connections(self) -> str:
        return self._t('node.editor.lbl.allowed', 'Allowed connections:')

    def cap_unlimited(self) -> str:
        return self._t('node.editor.cap.unlimited', 'unlimited (∞)')

    def cap_na(self) -> str:
        return self._t('node.editor.cap.na', 'n/a')

    def side_label(self, side: str) -> str:
        if side == "input":
            return self._t('node.editor.side.input', 'Input')
        if side == "output":
            return self._t('node.editor.side.output', 'Output')
        return side

    def hint_click_start(self) -> str:
        return self._t('node.editor.hint.click_start', 'Click: start a new connection')

    def hint_ctrl_click_rewire(self) -> str:
        return self._t('node.editor.hint.ctrl_rewire', 'Ctrl+Click: rewire/detach existing')

    def port_tooltip(self, node_name: str, side_label: str, prop_id: str, cap_str: str) -> str:
        return (
            f"{self.lbl_node()} {node_name}\n"
            f"{self.lbl_port()} {side_label} • {prop_id}\n"
            f"{self.lbl_allowed_connections()} {cap_str}\n\n"
            f"{self.hint_click_start()}\n"
            f"{self.hint_ctrl_click_rewire()}"
        )

    def port_capacity_tooltip(self, cap_text: str) -> str:
        return f"{self.lbl_allowed_connections()} (IN/OUT): {cap_text}"


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.05 20:00:00                  #
# ================================================== #

from typing import Any, Optional


def safe_get(obj: Any, path: str) -> Any:
    """
    Dot-path getter for dicts and objects.

    :param obj: Source object or dict
    :param path: Dot-separated path, e.g. 'a.b.0.c'
    :return: Value at path or None
    """
    cur = obj
    for seg in path.split("."):
        if cur is None:
            return None
        if isinstance(cur, dict):
            cur = cur.get(seg)
        else:
            if seg.isdigit() and isinstance(cur, (list, tuple)):
                idx = int(seg)
                if 0 <= idx < len(cur):
                    cur = cur[idx]
                else:
                    return None
            else:
                cur = getattr(cur, seg, None)
    return cur


def as_int(val: Any) -> Optional[int]:
    """
    Coerce to int if possible, else None.

    :param val: Input value
    :return: int or None
    """
    if val is None:
        return None
    try:
        return int(val)
    except Exception:
        try:
            return int(float(val))
        except Exception:
            return None


def capture_openai_usage(state, u_obj: Any):
    """
    Extract usage for OpenAI/xAI-compatible chunks.

    :param state: Chat state
    :param u_obj: Usage object/dict
    """
    if not u_obj:
        return
    state.usage_vendor = "openai"
    in_tok = as_int(safe_get(u_obj, "input_tokens")) or as_int(safe_get(u_obj, "prompt_tokens"))
    out_tok = as_int(safe_get(u_obj, "output_tokens")) or as_int(safe_get(u_obj, "completion_tokens"))
    total = as_int(safe_get(u_obj, "total_tokens"))
    reasoning = (
        as_int(safe_get(u_obj, "output_tokens_details.reasoning_tokens")) or
        as_int(safe_get(u_obj, "completion_tokens_details.reasoning_tokens")) or
        as_int(safe_get(u_obj, "reasoning_tokens")) or
        0
    )
    out_with_reason = (out_tok or 0) + (reasoning or 0)
    state.usage_payload = {"in": in_tok, "out": out_with_reason, "reasoning": reasoning or 0, "total": total}


def capture_google_usage(state, um_obj: Any):
    """
    Extract usage for Google python-genai; prefer total - prompt to include reasoning.

    :param state: Chat state
    :param um_obj: Usage metadata object/dict
    """
    if not um_obj:
        return
    state.usage_vendor = "google"
    prompt = (
        as_int(safe_get(um_obj, "prompt_token_count")) or
        as_int(safe_get(um_obj, "prompt_tokens")) or
        as_int(safe_get(um_obj, "input_tokens"))
    )
    total = (
        as_int(safe_get(um_obj, "total_token_count")) or
        as_int(safe_get(um_obj, "total_tokens"))
    )
    candidates = (
        as_int(safe_get(um_obj, "candidates_token_count")) or
        as_int(safe_get(um_obj, "output_tokens"))
    )
    reasoning = (
        as_int(safe_get(um_obj, "candidates_reasoning_token_count")) or
        as_int(safe_get(um_obj, "reasoning_tokens")) or 0
    )
    if total is not None and prompt is not None:
        out_total = max(0, total - prompt)
    else:
        out_total = candidates
    state.usage_payload = {"in": prompt, "out": out_total, "reasoning": reasoning or 0, "total": total}


def collect_google_citations(ctx, state, chunk: Any):
    """
    Collect web citations (URLs) from Google GenAI stream.

    :param ctx: Chat context
    :param state: Chat state
    :param chunk: Incoming streaming chunk
    """
    try:
        cands = getattr(chunk, "candidates", None) or []
    except Exception:
        cands = []

    if not isinstance(state.citations, list):
        state.citations = []

    def _add_url(url: Optional[str]):
        if not url or not isinstance(url, str):
            return
        url = url.strip()
        if not (url.startswith("http://") or url.startswith("https://")):
            return
        if ctx.urls is None:
            ctx.urls = []
        if url not in state.citations:
            state.citations.append(url)
        if url not in ctx.urls:
            ctx.urls.append(url)

    for cand in cands:
        gm = safe_get(cand, "grounding_metadata") or safe_get(cand, "groundingMetadata")
        if gm:
            atts = safe_get(gm, "grounding_attributions") or safe_get(gm, "groundingAttributions") or []
            try:
                for att in atts or []:
                    for path in (
                        "web.uri",
                        "web.url",
                        "source.web.uri",
                        "source.web.url",
                        "source.uri",
                        "source.url",
                        "uri",
                        "url",
                    ):
                        _add_url(safe_get(att, path))
            except Exception:
                pass
            for path in (
                "search_entry_point.uri",
                "search_entry_point.url",
                "searchEntryPoint.uri",
                "searchEntryPoint.url",
                "search_entry_point.rendered_content_uri",
                "searchEntryPoint.rendered_content_uri",
            ):
                _add_url(safe_get(gm, path))

        cm = safe_get(cand, "citation_metadata") or safe_get(cand, "citationMetadata")
        if cm:
            cit_arrays = (
                safe_get(cm, "citation_sources") or
                safe_get(cm, "citationSources") or
                safe_get(cm, "citations") or []
            )
            try:
                for cit in cit_arrays or []:
                    for path in ("uri", "url", "source.uri", "source.url", "web.uri", "web.url"):
                        _add_url(safe_get(cit, path))
            except Exception:
                pass

        try:
            parts = safe_get(cand, "content.parts") or []
            for p in parts:
                pcm = safe_get(p, "citation_metadata") or safe_get(p, "citationMetadata")
                if pcm:
                    arr = (
                        safe_get(pcm, "citation_sources") or
                        safe_get(pcm, "citationSources") or
                        safe_get(pcm, "citations") or []
                    )
                    for cit in arr or []:
                        for path in ("uri", "url", "source.uri", "source.url", "web.uri", "web.url"):
                            _add_url(safe_get(cit, path))
                gpa = safe_get(p, "grounding_attributions") or safe_get(p, "groundingAttributions") or []
                for att in gpa or []:
                    for path in ("web.uri", "web.url", "source.web.uri", "source.web.url", "uri", "url"):
                        _add_url(safe_get(att, path))
        except Exception:
            pass

    if state.citations and (ctx.urls is None or not ctx.urls):
        ctx.urls = list(state.citations)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.14 01:00:00                  #
# ================================================== #

from typing import Optional, Dict, Any
from pydantic import Field

from llama_index.core.workflow import (
    Event,
)

class StepEvent(Event):
    """Represents an event that occurs during a step in the workflow."""
    name: str # eg. "make_plan", "execute_plan", "subtask"
    index: Optional[int] = None
    total: Optional[int] = None
    meta: Dict[str, Any] = Field(default_factory=dict)